/**CHeader*********************************************************************

  FileName    [bddfsm.c]

  PackageName [fsm]

  Synopsis    [The new preliminary interface to the fsm package]

  Description [The new preliminary interface to the fsm package.]

  SeeAlso     [optional]

  Author      [Marco Pistore]

  Copyright   [This file is part of the ``simulator'' package of NuSMV
  version 2.  Copyright (C) 2003 ITC-irst and University of Trento.

  NuSMV version 2 is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2 of
  the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it> or
  email to <nusmv-users@irst.itc.it>.  Please report bugs to
  <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to
  <nusmv@irst.itc.it>.]

******************************************************************************/
#include "bddfsm.h"
#include "grammar.h"
#include "ddext.h"

extern FILE * nusmv_stderr;
extern DdManager * dd_manager;
extern options_ptr options;
extern bdd_ptr state_variables_bdd;

#define NOTIMPLEMENTED \
  fprintf(nusmv_stderr, "Function %s is not implemented.\n", \
          __PRETTY_FUNCTION__), \
  abort()

BddFsm BddFsm_Alloc()
{
  return Compile_FsmBddAlloc();
}

BddFsm BddFsm_Dup(BddFsm fsm)
{
  return Compile_FsmBddDup(fsm);
}

void BddFsm_Free(BddFsm fsm)
{
  Compile_FsmBddFree(fsm);
}

bdd_ptr BddFsm_GetInit(BddFsm fsm)
{
  return bdd_dup(Compile_FsmBddGetInit(fsm));
}

void BddFsm_SetInit(BddFsm fsm, bdd_ptr init)
{
  Compile_FsmBddSetInit(fsm,bdd_dup(init));
}

bdd_ptr BddFsm_GetInputs(BddFsm fsm)
{
  /** Modify **/
  return bdd_one(dd_manager);
}

void BddFsm_SetInputs(BddFsm fsm, bdd_ptr inputs)
{
  NOTIMPLEMENTED;
}

bdd_ptr BddFsm_GetFairStates(BddFsm fsm)
{
  /** Add reachable if computed */
  return bdd_dup(Compile_FsmBddGetFairStates(fsm));
}

void BddFsm_SetFairStates(BddFsm fsm, bdd_ptr states)
{
  Compile_FsmBddSetFairStates(fsm,bdd_dup(states));
}

bdd_ptr BddFsm_GetStates(BddFsm fsm)
{
  /** Add reachable if computed */
  return bdd_dup(Compile_FsmBddGetInvar(fsm));
}

void BddFsm_SetStates(BddFsm fsm, bdd_ptr states)
{
  Compile_FsmBddSetInvar(fsm,bdd_dup(states));
}

/** Cube */
bdd_ptr BddFsm_GetStateSupport(BddFsm fsm)
{
  extern bdd_ptr state_variables_bdd;
  return bdd_dup(state_variables_bdd);
}

/** Cube */
bdd_ptr BddFsm_GetNextStateSupport(BddFsm fsm)
{
  extern bdd_ptr next_state_variables_bdd;
  return bdd_dup(next_state_variables_bdd);
}

/** Cube */
bdd_ptr BddFsm_GetInputSupport(BddFsm fsm)
{
  extern bdd_ptr input_variables_bdd;
  return bdd_dup(input_variables_bdd);
}

/*** applicano maschere a bdd generici ***/

bdd_ptr BddFsm_GetMask(BddFsm fsm)
{
  /*** :HACK: Caching the computation of the mask ***/
  /*** :WARNING: The caching mechanism does not take
       into account updates of the fsm ***/
  static BddFsm cached_fsm = (BddFsm)NULL;
  static bdd_ptr cached_mask = (bdd_ptr)NULL;
  
  if (cached_fsm == fsm && cached_mask) return bdd_dup(cached_mask);
  if (cached_mask != (bdd_ptr)NULL) bdd_free(dd_manager, cached_mask);

  cached_fsm = fsm;

  cached_mask = bdd_one(dd_manager);
  {
    bdd_ptr state_mask = BddFsm_GetStateMask(fsm);
    bdd_ptr next_state_mask = BddFsm_GetNextStateMask(fsm);
    bdd_ptr input_mask = BddFsm_GetInputMask(fsm);

    bdd_and_accumulate(dd_manager,&cached_mask,state_mask);
    bdd_and_accumulate(dd_manager,&cached_mask,next_state_mask);
    bdd_and_accumulate(dd_manager,&cached_mask,input_mask);

    bdd_free(dd_manager, state_mask);
    bdd_free(dd_manager, next_state_mask);
    bdd_free(dd_manager, input_mask);
  }
  return bdd_dup(cached_mask);
}

bdd_ptr BddFsm_GetStateMask(BddFsm fsm)
{
  /*** :HACK: Caching the computation of the mask ***/
  /*** :WARNING: The caching mechanism does not take
       into account updates of the fsm ***/
  static BddFsm cached_fsm = (BddFsm)NULL;
  static bdd_ptr cached_mask = (bdd_ptr)NULL;
  
  if (cached_fsm == fsm && cached_mask) return bdd_dup(cached_mask);
  if (cached_mask != (bdd_ptr)NULL) bdd_free(dd_manager, cached_mask);

  cached_fsm = fsm;

  cached_mask = bdd_one(dd_manager);
  {
    node_ptr vars = BddFsm_GetStateVars(fsm);
    for(; vars != Nil; vars = cdr(vars)) {
      bdd_ptr mask = BddFsm_GetVarMask(fsm, car(vars));
      bdd_and_accumulate(dd_manager, &cached_mask, mask);
      bdd_free(dd_manager, mask);
    }
  }
  return bdd_dup(cached_mask);
}

bdd_ptr BddFsm_GetNextStateMask(BddFsm fsm)
{
  /*** :HACK: Caching the computation of the mask ***/
  /*** :WARNING: The caching mechanism does not take
       into account updates of the fsm ***/
  static BddFsm cached_fsm = (BddFsm)NULL;
  static bdd_ptr cached_mask = (bdd_ptr)NULL;
  
  if (cached_fsm == fsm && cached_mask) return bdd_dup(cached_mask);
  if (cached_mask != (bdd_ptr)NULL) bdd_free(dd_manager, cached_mask);

  cached_fsm = fsm;

  {
    bdd_ptr state_mask = BddFsm_GetStateMask(fsm);
    cached_mask = BddFsm_ShiftFwd(fsm, state_mask);
    bdd_free(dd_manager, state_mask);
  }

  return bdd_dup(cached_mask);
}

bdd_ptr BddFsm_GetInputMask(BddFsm fsm)
{
  /*** :HACK: Caching the computation of the mask ***/
  /*** :WARNING: The caching mechanism does not take
       into account updates of the fsm ***/
  static BddFsm cached_fsm = (BddFsm)NULL;
  static bdd_ptr cached_mask = (bdd_ptr)NULL;
  
  if (cached_fsm == fsm && cached_mask) return bdd_dup(cached_mask);
  if (cached_mask != (bdd_ptr)NULL) bdd_free(dd_manager, cached_mask);

  cached_fsm = fsm;

  cached_mask = bdd_one(dd_manager);
  {
    node_ptr vars = BddFsm_GetInputVars(fsm);
    for(; vars != Nil; vars = cdr(vars)) {
      bdd_ptr mask = BddFsm_GetVarMask(fsm, car(vars));
      bdd_and_accumulate(dd_manager, &cached_mask, mask);
      bdd_free(dd_manager, mask);
    }
  }
  return bdd_dup(cached_mask);
}

bdd_ptr BddFsm_Mask(BddFsm fsm, bdd_ptr bdd)
{
  bdd_ptr mask = BddFsm_GetMask(fsm);
  bdd_ptr result = bdd_and(dd_manager,bdd,mask);
  bdd_free(dd_manager,mask);
  return result;
}

bdd_ptr BddFsm_MaskStates(BddFsm fsm, bdd_ptr bdd)
{
  bdd_ptr mask = BddFsm_GetStateMask(fsm);
  bdd_ptr result = bdd_and(dd_manager,bdd,mask);
  bdd_free(dd_manager,mask);
  return result;
}

bdd_ptr BddFsm_MaskNextStates(BddFsm fsm, bdd_ptr bdd)
{
  bdd_ptr mask = BddFsm_GetNextStateMask(fsm);
  bdd_ptr result = bdd_and(dd_manager,bdd,mask);
  bdd_free(dd_manager,mask);
  return result;
}

bdd_ptr BddFsm_MaskInputs(BddFsm fsm, bdd_ptr bdd)
{
  bdd_ptr mask = BddFsm_GetInputMask(fsm);
  bdd_ptr result = bdd_and(dd_manager,bdd,mask);
  bdd_free(dd_manager,mask);
  return result;
}

node_ptr BddFsm_GetStateVars(BddFsm fsm)
{
  extern node_ptr state_variables;
  return state_variables;
}

node_ptr BddFsm_GetNextStateVars(BddFsm fsm)
{
  node_ptr state_vars = BddFsm_GetStateVars(fsm);
  node_ptr next_state_vars = Nil;
  while (state_vars) {
    next_state_vars = cons(find_node(NEXT,car(state_vars),Nil), next_state_vars);
    state_vars = cdr(state_vars);
  }
  return reverse(next_state_vars);
}

node_ptr BddFsm_GetInputVars(BddFsm fsm)
{
  extern node_ptr input_variables;
  return input_variables;
}

/*** restituisce cubo, range e encoding di variabile ***/
bdd_ptr BddFsm_GetVarSupport(BddFsm fsm, node_ptr var)
{
  add_ptr range_add;
  add_ptr support_add;
  bdd_ptr support_bdd;

  node_ptr q = lookup_symbol_hash(var);
  if (!q || node_get_type(q) != VAR) {
    fprintf(nusmv_stderr,"BddFsm_GetVarCube: wrong variable\n");
    abort();
  }
  range_add = (add_ptr)cdr(q);

  support_add = add_support(dd_manager, range_add);
  support_bdd = add_to_bdd(dd_manager, support_add);

  add_free(dd_manager, range_add);
  add_free(dd_manager, support_add);

  return support_bdd;
}

node_ptr BddFsm_GetVarRange(BddFsm fsm, node_ptr var)
{
  node_ptr q = lookup_symbol_hash(var);
  if (!q || node_get_type(q) != VAR) {
    fprintf(nusmv_stderr,"BddFsm_GetVarRange: wrong variable\n");
    abort();
  }
  return cdr(q);
}

add_ptr BddFsm_GetVarEncoding(BddFsm fsm, node_ptr var) 
{
  node_ptr q = lookup_symbol_hash(var);
  if (!q || node_get_type(q) != VAR) {
    fprintf(nusmv_stderr,"BddFsm_GetVarEncoding: wrong variable\n");
    abort();
  }
  return (add_ptr)car(q);
}

bdd_ptr BddFsm_GetVarMask(BddFsm fsm, node_ptr var)
{
  node_ptr q = lookup_symbol_hash(var);
  if (!q || node_get_type(q) != VAR) {
    fprintf(nusmv_stderr,"BddFsm_GetVarMask: wrong variable\n");
    abort();
  }
  {
    add_ptr var_add = (add_ptr)car(q);
    int bits;
#warning HARDCODED OFFSET 2
    add_ptr amask = make_var_mask(dd_manager, (bdd_ptr)var_add, &bits, 2);
    bdd_ptr mask = add_to_bdd(dd_manager, amask);
    bdd_free(dd_manager,amask);
    return mask;
  }
}

/*** shift current <-> next ***/
bdd_ptr BddFsm_ShiftFwd(BddFsm fsm, bdd_ptr states)
{
  return bdd_shift_forward(dd_manager, states);
}

bdd_ptr BddFsm_ShiftBwd(BddFsm fsm, bdd_ptr next_states)
{
  return bdd_shift_backward(dd_manager, next_states);
}

CPTrans BddFsm_GetCPTrans(BddFsm fsm)
{
  CPTrans_Ptr trans;
  switch(get_partition_method(options)) {
  case Monolithic:
    {
      trans = Compile_FsmBddGetMonoTrans(fsm);
      break;
    }
  case Threshold:
    {
      trans = Compile_FsmBddGetThreshold(fsm);
      break;
    }
  case Iwls95CP:
    {
      trans = Compile_FsmBddGetIwls95CP(fsm);
      break;
    }
  default:
    {
      rpterr("Bdd_GetCPTrans: partitioning method not available.");
      assert(0);
    }
  }
  return trans;
}

void BddFsm_SetCPTrans(BddFsm fsm, CPTrans cp)
{
  switch(get_partition_method(options)) {
  case Monolithic:
    {
      Compile_FsmBddSetMonoTrans(fsm,cp);
      break;
    }
  case Threshold:
    {
      Compile_FsmBddSetThreshold(fsm,cp);
      break;
    }
  case Iwls95CP:
    {
      Compile_FsmBddSetIwls95CP(fsm,cp);
      break;
    }
  default:
    {
      rpterr("Bdd_SetCPTrans: partitioning method not available.");
      assert(0);
    }
  }
}

/*** da evitare ***/
bdd_ptr BddFsm_GetTrans(BddFsm fsm)
{
  /*** :HACK: Caching the computation of the trans ***/
  /*** :WARNING: The caching mechanism does not take
       into account updates of the fsm ***/
  static BddFsm cached_fsm = (BddFsm)NULL;
  static bdd_ptr cached_trans = (bdd_ptr)NULL;

  CPTrans_Ptr trans;
  CPList fwd_trans;
  bdd_ptr result;

  if (cached_fsm == fsm && cached_trans) return bdd_dup(cached_trans);
  if (cached_trans != (bdd_ptr)NULL) bdd_free(dd_manager, cached_trans);

  trans = BddFsm_GetCPTrans(fsm);
  fwd_trans = CPTransGetForward(trans);
  
  result = bdd_one(dd_manager);
  while(CPListIsNotEmpty(fwd_trans)) {
    CPCluster * Ci = CPListGetItem(fwd_trans);
    bdd_and_accumulate(dd_manager, &result, CPClusterGetTi(Ci));
    fwd_trans = CPListNext(fwd_trans);
  }

  cached_fsm = fsm;
  cached_trans = bdd_dup(result);

  return(result);
}

/*** per model checking ***/
bdd_ptr BddFsm_ImageFwd(BddFsm fsm, bdd_ptr states)
{
  return Img_ImageFwd(fsm, states);
}

bdd_ptr BddFsm_ImageBwd(BddFsm fsm, bdd_ptr states)
{
  return Img_ImageBwd(fsm, states);
}

static bdd_ptr bddfsm_GetStatesInputsWithSuccessor(BddFsm fsm)
{
  /*** :HACK: Caching the computation of the si with successor ***/
  /*** :WARNING: The caching mechanism does not take
       into account updates of the fsm ***/
  static BddFsm cached_fsm = (BddFsm)NULL;
  static bdd_ptr cached_si = (bdd_ptr)NULL;

  if (cached_fsm == fsm && cached_si) return bdd_dup(cached_si);
  if (cached_si != (bdd_ptr)NULL) bdd_free(dd_manager, cached_si);

  {
    bdd_ptr one = bdd_one(dd_manager);
    cached_si = ex(fsm,one);
    bdd_free(dd_manager,one);
  }

  return bdd_dup(cached_si);
}

/*** per planning ***/
bdd_ptr BddFsm_StrongImageBwd(BddFsm fsm, bdd_ptr states)
{
  bdd_ptr not_states = bdd_not(dd_manager, states);
  bdd_ptr ex_not_states = ex(fsm,not_states);
  bdd_ptr ax_states = bdd_not(dd_manager, ex_not_states);

  bdd_ptr sa_with_successor = bddfsm_GetStatesInputsWithSuccessor(fsm);

  bdd_and_accumulate(dd_manager,&ax_states,sa_with_successor);

  bdd_free(dd_manager, not_states);
  bdd_free(dd_manager, ex_not_states);
  bdd_free(dd_manager, sa_with_successor);
  return(ax_states);
}

bdd_ptr BddFsm_WeakImageBwd(BddFsm fsm, bdd_ptr states)
{
  return ex(fsm,states);
}

static bdd_ptr bddfsm_ImgBwd_i(BddFsm fsm, bdd_ptr states, int i)
{
  node_ptr inode = find_node(NUMBER,(node_ptr)(i-1),Nil);

  bdd_ptr trans = BddFsm_GetTrans(fsm);
  add_ptr card_add = NULL;
  bdd_ptr wpi = NULL;

#warning FAIR STATES MISSING

  { 
    bdd_ptr invar = BddFsm_GetStates(fsm);
    bdd_ptr next_invar = BddFsm_ShiftFwd(fsm, invar);
    bdd_and_accumulate(dd_manager, &trans, next_invar);
    bdd_free(dd_manager, next_invar);
    bdd_free(dd_manager, invar);
  }
  
  {
    bdd_ptr next_states = BddFsm_ShiftFwd(fsm, states);
    bdd_ptr masked_next_states = BddFsm_MaskNextStates(fsm, next_states);
    bdd_and_accumulate(dd_manager, &trans, masked_next_states);
    bdd_free(dd_manager, masked_next_states);
    bdd_free(dd_manager, next_states);
  }

  {
    add_ptr trans_add = bdd_to_add(dd_manager, trans);
    bdd_ptr next_vars_bdd = BddFsm_GetNextStateSupport(fsm);
    add_ptr next_vars_add = bdd_to_add(dd_manager, next_vars_bdd);

    card_add = add_sum(dd_manager, trans_add,next_vars_add);

    add_free(dd_manager, trans_add);
    bdd_free(dd_manager, next_vars_bdd);
    add_free(dd_manager, next_vars_add);
  }

  wpi = add_bdd_strictthreshold(dd_manager, card_add, inode);

  bdd_free(dd_manager,trans);
  add_free(dd_manager,card_add);

  return wpi;
}

static bdd_ptr bddfsm_MImgBwdAux(BddFsm fsm, 
				 bdd_ptr_set exs, 
				 bdd_ptr ex_states, 
				 int card) 
{
  if (exs) {
    bdd_ptr ex_states_ext = bdd_or(dd_manager, ex_states, (bdd_ptr)car(exs));
    bdd_ptr ex0 = bddfsm_MImgBwdAux(fsm,cdr(exs),ex_states,card);
    bdd_ptr ex1 = bddfsm_MImgBwdAux(fsm,cdr(exs),ex_states_ext,card+1);
    bdd_ptr ex = bdd_and(dd_manager, ex0, ex1);
    bdd_free(dd_manager,ex_states_ext);
    bdd_free(dd_manager,ex0);
    bdd_free(dd_manager,ex1);
    return ex;
  } else {
    return bddfsm_ImgBwd_i(fsm,ex_states,card);
  }
}

bdd_ptr BddFsm_MultipleImageBwd(BddFsm fsm, bdd_ptr_set states_set)
{
  if(states_set == Nil) {
    return BddFsm_WeakImageBwd(fsm, bdd_one(dd_manager));
  }
  if(cdr(states_set) == Nil) {
    return BddFsm_WeakImageBwd(fsm, (bdd_ptr)car(states_set));
  }
  return bddfsm_MImgBwdAux(fsm,states_set,bdd_zero(dd_manager),0);
}

/*** conformant images ***/
static bdd_ptr conformant_executable(BddFsm fsm, bdd_ptr bs)
{
  bdd_ptr tmp, result;
  bdd_ptr one = bdd_one(dd_manager);
  bdd_ptr states_with_successor = ex(fsm, one);
  bdd_ptr not_bimg = bdd_not(dd_manager, states_with_successor);
  bdd_ptr vars_bdd = BddFsm_GetStateSupport(fsm);

  tmp = bdd_and_abstract(dd_manager, bs, not_bimg, vars_bdd);
  result = bdd_not(dd_manager, tmp);
  bdd_free(dd_manager, tmp);
  bdd_free(dd_manager, vars_bdd); 
  bdd_free(dd_manager, not_bimg);
  bdd_free(dd_manager, states_with_successor);
  bdd_free(dd_manager, one);
  return result;
}

/*
  the forward BS expansion does not need to quantify out input
  variables. The following code handles this "feature" by hacking the
  image code to show the desired behavior.
*/
static BddFsm static_bdd_fsm = (BddFsm)NULL;
static BddFsm old_static_bdd_fsm = (BddFsm)NULL;

static BddFsm make_static_fsm(BddFsm fsm) {
  CPList fwd = NULL;

  if (old_static_bdd_fsm == fsm) return(static_bdd_fsm);
  old_static_bdd_fsm = fsm;
  if (static_bdd_fsm != (BddFsm)NULL) Compile_FsmBddFree(static_bdd_fsm);
  static_bdd_fsm = Compile_FsmBddDup(fsm);
  if (Compile_FsmBddGetMonoTrans(static_bdd_fsm) != (CPTrans_Ptr)NULL) {
    fwd = CPTransGetForward(Compile_FsmBddGetMonoTrans(static_bdd_fsm));
  }
  if (Compile_FsmBddGetIwls95CP(static_bdd_fsm) != (CPTrans_Ptr)NULL) {
    fwd = CPTransGetForward(Compile_FsmBddGetIwls95CP(static_bdd_fsm));
  }
  if (Compile_FsmBddGetThreshold(static_bdd_fsm) != (CPTrans_Ptr)NULL) {
    fwd = CPTransGetForward(Compile_FsmBddGetThreshold(static_bdd_fsm));
  }
  /* side effects on the fwd structure */
  CPListBuildSchedule(fwd, BddFsm_GetStateSupport(fsm));
  /* CPTransSetForward(Compile_FsmBddGetThreshold(fsm), fwd); */
  return(static_bdd_fsm);
}

/**Function********************************************************************

  Synopsis [Performs both forward or backward image computation.]

  Description [Depending the Cluster list passed as argument, this
  function computes respectively the forward or backward image of a
  given state. CP version.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static bdd_ptr CPImageBoth(DdManager *dd, CPList Clist, bdd_ptr S, bdd_ptr vars)
{
  bdd_ptr result;
  int it = 0;

  bdd_ptr cur_prod = bdd_dup(S);
  long maxSize = 0;
  long intermediateSize = 0;
  
  while(CPListIsNotEmpty(Clist)) {

    {
      CPCluster * Ci = CPListGetItem(Clist);
      
      bdd_ptr new_p = bdd_and_abstract(dd, cur_prod, CPClusterGetTi(Ci), CPClusterGetEi(Ci));

      it++;

     if (get_verbose_level(options) >= 1) { 
	intermediateSize = bdd_size(dd, new_p);
	(void) fprintf(nusmv_stdout, "          Size of intermediate product = %10ld (BDD nodes).\n", intermediateSize);
	if (maxSize < intermediateSize) {
	  maxSize = intermediateSize;
	}
     }

      bdd_free(dd, cur_prod);
      cur_prod = new_p;
      Clist = CPListNext(Clist);
    }

  }
  result = bdd_forsome(dd, cur_prod, vars);
  bdd_free(dd, cur_prod);

  if (get_verbose_level(options) >= 1){
    (void) fprintf(nusmv_stdout, "                 Total size of product = %10ld (BDD nodes).\n", (long)bdd_size(dd,result));
    (void) fprintf(nusmv_stdout, "Max. BDD size for intermediate product = %10ld (BDD nodes)\n", maxSize);
    fprintf(nusmv_stdout, "Variables number in intermediate product = %10d (BDD nodes).\n", bdd_size(dd_manager,bdd_support(dd_manager,result)));
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Computes the forward image of a set of  states.]

  Description        [Computes the forward image of a set of states.
  It uses an implicitly conjoined transition relation. The resulting
  set is in term of current state variables.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static bdd_ptr ImgImageFwdCPNI(CPTrans_Ptr trans, bdd_ptr g)
{
  node_ptr CP_Fwd_trans = CPTransGetForward(trans);
  bdd_ptr vars = bdd_dup(state_variables_bdd);
  bdd_ptr tmp = CPImageBoth(dd_manager, CP_Fwd_trans, g, vars);
  bdd_ptr result = bdd_shift_backward(dd_manager, tmp);
  bdd_free(dd_manager, tmp);
  bdd_free(dd_manager, vars);
  return(result);
}

/**Function********************************************************************

  Synopsis           [Computes the forward image of a set of states.]

  Description        [Computes the forward image of a set of states.
  The resulting set is in term of current state variables. Depending
  the partitioning heuristich choosen, the corresponding image
  function is used.]

  SideEffects        [None]

  SeeAlso            [Img_ImageBwd]

******************************************************************************/
bdd_ptr Img_ImageFwdNI(Fsm_BddPtr fsm, bdd_ptr g)
{
  Partition_Method part;

  part = get_partition_method(options);
  switch(part) {
  case Monolithic:
    {
      return(ImgImageFwdCPNI(Compile_FsmBddGetMonoTrans(fsm), g));
    }
  case Threshold:
    {
      return(ImgImageFwdCPNI(Compile_FsmBddGetThreshold(fsm), g));
    }
  case Iwls95CP:
    {
      return(ImgImageFwdCPNI(Compile_FsmBddGetIwls95CP(fsm), g));
    }
  default:
    {
      rpterr("Img_ImageFwdNI: partitioning method not available.");
      return((bdd_ptr)Nil);
    }
  }
}

bdd_ptr BddFsm_BsImageFwd(BddFsm fsm, bdd_ptr bs)
{
  bdd_ptr result;
  bdd_ptr executable = conformant_executable(fsm, bs);
  BddFsm static_fsm = make_static_fsm(fsm);

  bdd_and_accumulate(dd_manager, &executable, bs);
  result = Img_ImageFwdNI(static_fsm, executable);
  bdd_free(dd_manager, executable);
  return(result);
}

static BddFsm old_conf_fsm = (BddFsm)NULL;
static bdd_ptr old_states_with_successor = (bdd_ptr)NULL;

bdd_ptr BddFsm_BsImageBwd(BddFsm fsm, bdd_ptr bs)
{
  /* 
     forall x'. (R(x,an,x') \imp SP(x',a{n-1},...,a0)) = 
     \neg \exists x' \neg (\neg R(x,an,x') \or SP(x',a{n-1},...,a0)) =
     \neg \exists x' (R(x,an,x') \and \neg SP(x',a{n-1},...,a0)) =
  */
  bdd_ptr result;
  bdd_ptr image;
  bdd_ptr states_with_successor;
  bdd_ptr one = bdd_one(dd_manager);
  bdd_ptr not_bs = bdd_not(dd_manager, bs);

  if (fsm == old_conf_fsm) {
    states_with_successor = bdd_dup(old_states_with_successor);
  }
  else {
    if (old_states_with_successor != (bdd_ptr)NULL) bdd_free(dd_manager, old_states_with_successor);
    states_with_successor = ex(fsm, one);
    old_states_with_successor = bdd_dup(states_with_successor);
    old_conf_fsm = fsm;
  }

  image = ex(fsm, not_bs);
  result = bdd_not(dd_manager, image);
  bdd_and_accumulate(dd_manager, &result, states_with_successor);
  bdd_free(dd_manager, image);
  bdd_free(dd_manager, not_bs);
  bdd_free(dd_manager, states_with_successor);
  bdd_free(dd_manager, one);
  return(result);
}

int BddFsm_GetNumBddStateVars(BddFsm fsm) {
  bdd_ptr a = BddFsm_GetStateSupport(fsm);

  bdd_free(dd_manager, a);
  /* the terminal node must be avoided */
  return((int)(bdd_size(dd_manager, a) -1));
}

int BddFsm_GetStateLevel(BddFsm fsm)  {
  bdd_ptr a = BddFsm_GetStateSupport(fsm);

  bdd_free(dd_manager, a);
  return((int)(bdd_readperm(dd_manager, a)-1));
}


/*** expr -> [ab]dd ***/
bdd_ptr BddFsm_EvalCond(BddFsm fsm, node_ptr cond)
{
  add_ptr expr_add = eval(cond, Nil);
  bdd_ptr result = add_to_bdd(dd_manager, expr_add);
  add_free(dd_manager,expr_add);
  return result;
}

add_ptr BddFsm_EvalExpr(BddFsm fsm, node_ptr expr)
{
  return eval(expr, Nil);
}

static node_ptr mk_true() 
{
  return find_node(TRUEEXP, Nil, Nil);
}

static node_ptr mk_false() 
{
  return find_node(FALSEEXP, Nil, Nil);
}

static node_ptr mk_and(node_ptr e1, node_ptr e2)
{
  if (node_get_type(e1) == FALSEEXP || node_get_type(e2) == FALSEEXP)
    return mk_false();
  if (node_get_type(e1) == TRUEEXP)
    return e2;
  if (node_get_type(e2) == TRUEEXP)
    return e1;
  return find_node(AND, e1, e2);
}

static node_ptr mk_or(node_ptr e1, node_ptr e2)
{
  if (node_get_type(e1) == TRUEEXP || node_get_type(e2) == TRUEEXP)
    return mk_true();
  if (node_get_type(e1) == FALSEEXP)
    return e2;
  if (node_get_type(e2) == FALSEEXP)
    return e1;
  return find_node(OR, e1, e2);
}

static node_ptr mk_eq(node_ptr e1, node_ptr e2)
{
  return find_node(EQUAL, e1, e2);
}

static node_ptr gen_multi_cond_term_aux(bdd_ptr conds, 
					node_ptr assign_list, 
					node_ptr var_list)
{
  if (var_list == Nil) {
    node_ptr ass = mk_true();
    while(assign_list) {
      node_ptr assign_el = car(assign_list);
      assign_list = cdr(assign_list);
      ass = mk_and(mk_eq(car(assign_el),cdr(assign_el)),ass);
    }
    return ass;
  }
  else {
    node_ptr var = car(var_list);
    node_ptr cdr_var_list = cdr(var_list);
    add_ptr var_add = eval(var,Nil);
    bdd_ptr var_supp = add_to_bdd(dd_manager,add_support(dd_manager,var_add));
    bdd_ptr conds_supp = bdd_support(dd_manager,conds);
    
    if(0) {
      fprintf(stdout,"Considering ");
      print_node(stdout,var);
      fprintf(stdout,"\n");
    }

    if (bdd_is_one(dd_manager,bdd_cube_intersection(dd_manager, var_supp, conds_supp))) {
      return gen_multi_cond_term_aux(conds, assign_list, cdr_var_list);
    }
    else {
      node_ptr ass = mk_false();
      node_ptr range;
      if (node_get_type(var) == NEXT) {
	range = cdr(lookup_symbol_hash(car(var)));
      } else {
	range = cdr(lookup_symbol_hash(var));
      }
      while(range) {
	node_ptr val = car(range);
	add_ptr eq_val_a = eval(mk_eq(var,val),Nil);
	bdd_ptr eq_val = add_to_bdd(dd_manager, eq_val_a);
	bdd_ptr new_conds = bdd_and(dd_manager, conds, eq_val);

	if (! bdd_is_zero(dd_manager, new_conds)) {
	  ass = mk_or(gen_multi_cond_term_aux(new_conds, 
					      cons(cons(var,val),assign_list), 
					      cdr_var_list),
		      ass);
	}
	range = cdr(range);
      }
      return ass;
    }
  }
}

static node_ptr gen_multi_cond_term(BddFsm fsm, bdd_ptr conds) 
{
  node_ptr all_vars = Nil;
  node_ptr t;

  if (bdd_is_one(dd_manager, conds)) {
    return(mk_true());
  }
  if (bdd_is_zero(dd_manager, conds)) {
    return(mk_false());
  }

  /* CURRENT STATE VARIABLES FIRST */
  for (t = BddFsm_GetStateVars(fsm); t; t = cdr(t)) {
    all_vars = cons(car(t), all_vars);
  }

  /* ACTION VARIABLES NEXT */
  for (t = BddFsm_GetInputVars(fsm); t; t = cdr(t)) {
    all_vars = cons(car(t), all_vars);
  }
  
  /* NEXT STATE VARIABLES LAST */
  for (t = BddFsm_GetNextStateVars(fsm); t; t = cdr(t)) {
    all_vars = cons(car(t), all_vars);
  }

  all_vars = reverse(all_vars);
  return gen_multi_cond_term_aux(conds, Nil, all_vars);
}

/*** [ab]dd -> expr ***/
node_ptr BddFsm_WffCond(BddFsm fsm, bdd_ptr bdd)
{
  return gen_multi_cond_term(fsm, bdd);
}

BddFsm save_fsm = NULL;
add_ptr save_add = NULL;
node_ptr save_expr = Nil;

static void bddfsm_wffexprval(node_ptr val)
{
  add_ptr val_add = add_leaf(dd_manager, val);
  if (!add_is_zero(dd_manager, val_add)) {
    add_ptr eq_add = add_equal(dd_manager, val_add, save_add);
    bdd_ptr eq_bdd = add_to_bdd(dd_manager, eq_add);
    node_ptr expr = find_node(COLON, BddFsm_WffCond(save_fsm, eq_bdd), val);
    save_expr = find_node(CASE,expr,save_expr);
    bdd_free(dd_manager, eq_bdd);
    add_free(dd_manager, eq_add);
  }
  add_free(dd_manager, val_add);
}

node_ptr BddFsm_WffExpr(BddFsm fsm, add_ptr add)
{
  save_add = bdd_dup(add);
  save_fsm = fsm;
  save_expr = find_node(FALSEEXP,Nil,Nil);
  add_walkleaves(bddfsm_wffexprval, add);
  bdd_free(dd_manager,save_add);
  save_add = NULL;
  save_fsm = NULL;
  {
    node_ptr result = save_expr;
    save_expr = Nil;
    return result;
  }
}

