/**CFile***********************************************************************

  FileName    [opts.c]

  PackageName [opts]

  Synopsis    [Generic handler of options. ]

  Description [Generic handler of options. An option is uniquely
  identified in an option handler by:
  <ul>
    <li> a name (it is the string of its name). 
    <li> a default value (it is the string of its name). 
    <li> a value (it is the string of its name). 
  </ul>
  When registering an option the user must specify two functions. The
  first is responsible of checking that the value passed while setting
  a value is a valid value for the given option. The second function
  is a function that transforms the stroed value in a value suitable
  to be used in the calling program. <br>

  For boolean options are provided special methods to register the
  option and for setting and getting a value associated to it.<br>

  For enumerative options are provided special methods to register the
  option and for setting and getting a value associated to it. An
  enumertive option is registered by providing an array of structures
  of type Opts_EnumRec. Similarly to the below declaration:<br>
  <pre>
    typedef enum {foo1, ...., fooN} fooenumtype;
    Opts_EnumRec foo[] = {"foo1", foo1, 
                          "foo2", foo2, 
                          ....
                          "fooN", fooN};
   ....
   handler = Opts_HandlerAlloc();
   Opts_HandlerRegisterEnumOption(handler, "foooption", "foo1", foo, N);

  if (Opts_HandlerGetEnumOptionValue(handler, "foooption") == foo2) {
     ...
  }

  ...

  switch(Opts_HandlerGetEnumOptionValue(handler, "foooption")) {
    case foo1:
    ...
    case fooN:
    ...
    default:
    ...
  }
  </pre>
  ]

  Author      [Marco Roveri]

  Copyright   [
  This file is part of the ``simulator'' package of NuSMV version 2.
  Copyright (C) 2003 by ITC-irst.

  NuSMV version 2 is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>.
  ]

*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include "opts.h" 
#include "st.h"

static char rcsid[] UTIL_UNUSED = "$Id: opts.c,v 1.3 2003/10/09 10:07:54 nusmv Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/
typedef enum st_retval opts_st_retval;
typedef struct _option_structure opt_rec;
typedef struct _option_structure * opt_ptr;
typedef struct _option_value_list ovl_rec;
typedef struct _option_value_list * ovl_ptr;
typedef boolean (* Opts_CheckOplFnType)(const char *, ovl_ptr);
typedef void * (* Opts_GetOplFnType)(const char *, ovl_ptr);

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

struct _Opts_HandlerRec {
  st_table * table;
  st_generator * gen;
};

struct _option_structure {
  char * name;
  char * default_value;
  char * value;
  ovl_ptr possible_values;
  Opts_CheckFnType check;
  Opts_ReturnFnType getvalue;
};

struct _option_value_list {
  ovl_ptr next;
  char * values;
  int valuee;
};

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

extern FILE * nusmv_stderr;

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

#define OPTS_STR_CMP_FUNC strcmp
#define OPTS_BOOLEAN_REC {{OPTS_FALSE_VALUE, false},{OPTS_TRUE_VALUE, true}}
#define OPTS_DELETE_ENTRY ST_DELETE

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static boolean opts_isinteger ARGS((const char *name));
static void * opts_getinteger ARGS((const char *value));

static opts_st_retval opts_hash_free ARGS((char *key, char *data, char * arg));
static opt_ptr option_alloc ARGS((void)); 
static void option_free ARGS((opt_ptr * o)); 
static char * opts_strsav ARGS((const char *s)); 
static opt_ptr option_create ARGS(( const char * name, const char * default_value, const char * value, ovl_ptr opl, Opts_CheckFnType check, Opts_ReturnFnType getvalue)); 
static boolean opt_enum_check ARGS((const char * value, ovl_ptr l)); 
static void * opt_enum_get ARGS((const char * value, ovl_ptr l)); 
/* static void * opt_bool_get ARGS((const char * value)); */
static ovl_ptr ovl_rec_alloc ARGS((void)); 
static ovl_ptr ovl_create_empty ARGS((void)); 
static int ovl_isempty ARGS((ovl_ptr l)); 
static int ovl_isnotempty ARGS((ovl_ptr l)); 
static ovl_ptr ovl_create ARGS((const char * value, int valuee)); 
static char * ovl_get_values ARGS((ovl_ptr l)); 
static int ovl_get_valuee ARGS((ovl_ptr l)); 
static ovl_ptr ovl_get_next ARGS((ovl_ptr l)); 
static ovl_ptr ovl_set_next ARGS((ovl_ptr l, ovl_ptr n)); 
static void ovl_free ARGS((ovl_ptr *l)); 
static int ovl_ispresent ARGS((ovl_ptr l, const char *value)); 
static int check_boolean ARGS((ovl_ptr l)); 

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/
/**Function********************************************************************

  Synopsis           [Creates an empty option handler]

  Description        [Creates an empty option handler. If it fails
  then it returns OPTS_ERROR.]

  SideEffects        [None]

  SeeAlso            [Opts_HandlerFree]

******************************************************************************/
Opts_HandlerPtr Opts_HandlerAlloc(void) {
  Opts_HandlerPtr result;

  result = ALLOC(Opts_HandlerRec, 1);
  if (result == (Opts_HandlerPtr)NULL) {
    fprintf(nusmv_stderr, "Opts_HandlerAlloc: Unable to allocate a generic option handler.\n");
    result = OPTS_ERROR;
  }
  else {
    st_table * h = st_init_table(OPTS_STR_CMP_FUNC, st_strhash);

    if (h == NIL(st_table)) {
      FREE(result);
      result = OPTS_ERROR;
    }
    else {
      result->table = h;
      result->gen = (st_generator *)NULL;
    }
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Frees an option handler.]

  Description        [Frees an option handler.]

  SideEffects        [None]

  SeeAlso            [Opts_HandlerAlloc]

******************************************************************************/
void Opts_HandlerFree(Opts_HandlerPtr *h) {
  assert((*h != (Opts_HandlerPtr)NULL) && (*h != OPTS_ERROR));

  st_foreach((*h)->table, opts_hash_free, (char *)NULL);
  st_free_table((*h)->table);
  FREE(*h);
}                       

/**Function********************************************************************

  Synopsis           [Registers an option in an option handler.]

  Description        [Registers an option in an option handler. Returns
  true if the registration of the option succeeds, false otherwise.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerRegisterOption(
 Opts_HandlerPtr h, 
 const char * name,
 const char * def,
 Opts_CheckFnType check,
 Opts_ReturnFnType get)
{
  opt_ptr opt;
  boolean result = false;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));
  if (st_lookup(h->table, (char *)name, (char **)&opt) == 0) {
    ovl_ptr ovl = ovl_create_empty();

    opt = option_create(name, def, def, ovl, check, get);
    if (opt != (opt_ptr)NULL) {
      if (st_insert(h->table, (char *)name, (char *)opt) != ST_OUT_OF_MEM) {
        result = true;
      }
    }
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Registers an enumerative option in an option handler.]

  Description        [Registers an enumerative option in an option
  handler. The possible values are stored in an array of strings given
  in input. The user is not required to provide any function to check
  and return a value. Returns true if the registration of the option succeeds,
  false otherwise.] 

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerRegisterEnumOption(
 Opts_HandlerPtr h,
 const char * name,
 const char * def,
 Opts_EnumRec pv[],
 int npv)
{
  int n;
  opt_ptr opt;
  ovl_ptr ovl;
  boolean result = false;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));
  if (st_lookup(h->table, (char *)name, (char **)&opt) == 0) {
    ovl = ovl_create_empty();
    for (n = 0; n < npv; n++) {
      ovl_ptr l = ovl_create(opts_strsav(pv[n].v), pv[n].e);

      ovl_set_next(l, ovl);
      ovl = l;
    }
    opt = option_create(name, def, def, ovl,
                        (Opts_CheckFnType)opt_enum_check,
                        (Opts_ReturnFnType)opt_enum_get);
    if (opt != (opt_ptr)NULL) {
      if (st_insert(h->table, (char *)name, (char *)opt) != ST_OUT_OF_MEM) {
        result = true;
      }
    }
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Registers a boolean option in an option handler.]

  Description        [Registers a boolean option in an option
  handler. The user is not required to provide any function to check
  and return a value. Returns true if the registration of the option succeeds,
  false otherwise.] 

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerRegisterBoolOption(
 Opts_HandlerPtr h, 
 const char * name,
 boolean value)
{
  boolean result = false;
  char * def;
  Opts_EnumRec pv[2] = OPTS_BOOLEAN_REC;
  
  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));

  def = (value == true) ? OPTS_TRUE_VALUE : OPTS_FALSE_VALUE;
  result = Opts_HandlerRegisterEnumOption(h, name, def, pv, 2);
  return(result);
}

/**Function********************************************************************

  Synopsis           [Registers an integer option in an option handler.]

  Description        [Registers an integer option in an option
  handler. The user is not required to provide any function to check
  and return a value. Returns true if the registration of the option succeeds,
  false otherwise.] 

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerRegisterIntOption(
 Opts_HandlerPtr h, 
 const char * name,
 int value)
{
  boolean result = false;
  char def[100];

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));
  sprintf(def, "%d", value);
  result = Opts_HandlerRegisterOption(h, name, def, opts_isinteger, opts_getinteger);
  return(result);
}

/**Function********************************************************************

  Synopsis           [Checks if an option has already been registered.]

  Description        [Checks if an option has already been
  registered. Returns true if it has been already registered false otherwise.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerIsOptionRegistered(Opts_HandlerPtr h, const char * name)
{
  opt_ptr opt;
  boolean result = false;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));
  if (st_lookup(h->table, (char *)name, (char **)&opt) == 1) {
    result = true;
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Checks if an option has already been registered.]

  Description        [Checks if an option has already been
  registered. Returns false if it has been already registered true otherwise.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerIsOptionNotRegistered(Opts_HandlerPtr h, const char * name)
{
  opt_ptr opt;
  boolean result = true;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));
  if (st_lookup(h->table, (char *)name, (char **)&opt) == 0) {
    result = false;
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Unregisters an option in an option handler.]

  Description        [Unregisters an option in an option handler. Returns
  true if the unregistration of the option succeeds, false otherwise.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerUnregisterOption(Opts_HandlerPtr h, const char * name)
{
  opt_ptr opt;
  boolean result = false;
  char * n = (char *)name;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));
  if (st_delete(h->table, &n, (char **)&opt) == 1) {
    assert(opt != (opt_ptr)NULL);
    option_free(&opt);
    result = true;
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Assigns the given value to a registered option.]

  Description        [Assigns the given value to an option registered
  in an option handler. Returns true if the setting of the value
  succeeds, false if the option name is not registered in the option
  handler or if the value to assigns does not is of the type allowed
  for the option.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerSetOptionValue(
 Opts_HandlerPtr h,
 const char * name,
 const char * value)
{
  opt_ptr opt;
  boolean result = false;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));

  if (st_lookup(h->table, (char *)name, (char **)&opt) == 1) {
    assert(opt != (opt_ptr)NULL);

    if (ovl_isempty(opt->possible_values) == 1) {
      /* user defined option */
      if ( (* opt->check)((char *)value) == true ) {
        if (opt->value != (char *)NULL) {
          FREE(opt->value);
        }
        opt->value = opts_strsav(value);
        result = true;
      }
    }
    else {
      Opts_CheckOplFnType f;

      f = (Opts_CheckOplFnType)opt->check;
      /* internally handled options */
      if ( (*f)(value, opt->possible_values) == true) {
        if (opt->value != (char *)NULL) {
          FREE(opt->value);
        }
        opt->value = opts_strsav(value);
        result = true;
      }
    }
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Assigns the given value to a registered option.]

  Description        [Assigns the given value to an option registered
  in an option handler. Returns true if the setting of the value
  succeeds, false if the option name is not registered in the option
  handler or if the value to assigns does not is of the type allowed
  for the option.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerSetEnumOptionValue(
 Opts_HandlerPtr h,
 const char * name,
 const char * value)
{
  return(Opts_HandlerSetOptionValue(h, name, value));
}

/**Function********************************************************************

  Synopsis           [Assigns the given value to a registered option.]

  Description        [Assigns the given value to an option registered
  in an option handler. Returns true if the setting of the value
  succeeds, false if the option name is not registered in the option
  handler or if the value to assigns does not is of the type allowed
  for the option.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerSetBoolOptionValue(
 Opts_HandlerPtr h,
 const char * name,
 boolean value)
{
  char * v;
  opt_ptr opt;
  boolean result = false;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));

  if (st_lookup(h->table, (char *)name, (char **)&opt) == 1) {
    assert(opt != (opt_ptr)NULL);
    assert(check_boolean(opt->possible_values) == 1);

    v = (value == true) ? OPTS_TRUE_VALUE : OPTS_FALSE_VALUE;
    result = true;
    if (opt->value != (char *)NULL) {
      FREE(opt->value);
    }
    opt->value = opts_strsav(v);
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Assigns the given value to a registered option.]

  Description        [Assigns the given value to an option registered
  in an option handler. Returns true if the setting of the value
  succeeds, false if the option name is not registered in the option
  handler or if the value to assigns does not is of the type allowed
  for the option.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerSetIntOptionValue(
 Opts_HandlerPtr h,
 const char * name,
 int value)
{
  char val[100];

  return(Opts_HandlerSetOptionValue(h, name, val));
}

/**Function********************************************************************

  Synopsis           [Assigns the default value to a registered option.]

  Description        [Assigns the default value to an option registered
  in an option handler. Returns true if the setting of the value
  succeeds, false if the option name is not registered in the option
  handler.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerResetOptionValue(Opts_HandlerPtr h, const char * name)
{
  opt_ptr opt;
  boolean result = false;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));

  if (st_lookup(h->table, (char *)name, (char **)&opt) == 1) {
    assert(opt != (opt_ptr)NULL);

    if (opt->value != (char *)NULL) {
      FREE(opt->value);
    }
    assert(opt->default_value != (char *)NULL);
    opt->value = opts_strsav(opt->default_value);
    result = true;
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Returns the value of a registered option.]

  Description        [Returns the value of an option registered
  in an option. OPTS_VALUE_ERROR is returned if the option is not a
  registered option.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
void * Opts_HandlerGetOptionValue(Opts_HandlerPtr h, const char * name)
{
  opt_ptr opt;
  void * result = OPTS_VALUE_ERROR;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));

  if (st_lookup(h->table, (char *)name, (char **)&opt) == 1) {
    assert(opt != (opt_ptr)NULL);

    if (opt->value != (char *)NULL) {
      if (ovl_isempty(opt->possible_values) == 1) {
        /* user defined option */
        result = (* opt->getvalue)(opt->value);
      }
      else {
        /* internally handled option */
        Opts_GetOplFnType f;

        f = (Opts_GetOplFnType)opt->getvalue;
        result = (*f)(opt->value, opt->possible_values);
      }
    }
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Returns the value of an enum option.]

  Description        [Returns the value of an enum option
  value. OPTS_VALUE_ERROR is returned if the option is not a
  registered option.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
int Opts_HandlerGetEnumOptionValue(Opts_HandlerPtr h, const char * name)
{
  return((int)Opts_HandlerGetOptionValue(h, name));
}

/**Function********************************************************************

  Synopsis           [Returns the value of a boolean option.]

  Description        [Returns the value of a boolean option
  value. OPTS_VALUE_ERROR is returned if the option is not a
  registered option.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
boolean Opts_HandlerGetBoolOptionValue(Opts_HandlerPtr h, const char * name)
{
  opt_ptr opt;
  boolean result = (boolean)OPTS_VALUE_ERROR;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));

  if (st_lookup(h->table, (char *)name, (char **)&opt) == 1) {
    assert(opt != (opt_ptr)NULL);
    assert(check_boolean(opt->possible_values) == 1);

    if (opt->value != (char *)NULL) {
        Opts_GetOplFnType f;

        f = (Opts_GetOplFnType)opt->getvalue;
        result = (boolean)(*f)(opt->value, opt->possible_values);
    }
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Returns the value of an int option.]

  Description        [Returns the value of an enum option
  value. OPTS_VALUE_ERROR is returned if the option is not a
  registered option.]

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
int Opts_HandlerGetIntOptionValue(Opts_HandlerPtr h, const char * name)
{
  return((int)Opts_HandlerGetOptionValue(h, name));
}

/**Function********************************************************************

  Synopsis           [Initializes a generator for an option handler.]

  Description        [Initializes a generator handler which when used
  with Opts_Gen() will progressively return each (name, value)
  record in the option handler.]

  SideEffects        [None]

  SeeAlso            [Opts_Gen Opts_GenFree]

******************************************************************************/
void Opts_GenInit(Opts_HandlerPtr h) {
  st_generator * gen;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));
  assert(h->gen == (st_generator *)NULL);

  gen = st_init_gen(h->table);
  if (gen != (st_generator *)NULL) {
    h->gen = gen;
  }
  else {
    fprintf(nusmv_stderr, "Opts_GenInit: Unable to allocate generator\n");
  }
}

/**Function********************************************************************

  Synopsis           [Gets the next pair (name, value) for an option handler.]

  Description        [Given a generator created by Opts_GenInit(),
     this routine returns the next (name, value) pair in the
     generation sequence. When there are no more items in the
     generation sequence,  the routine returns 0.]

  SideEffects        [None]

  SeeAlso            [Opts_GenInit Opts_GenFree]

******************************************************************************/
int Opts_Gen(Opts_HandlerPtr h, char **name, char **value)
{
  char * n;
  opt_ptr opt = (opt_ptr)1;
  int result;

  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));
  assert(h->gen != (st_generator *)NULL);

  *value = (char *)NULL;
  result = st_gen(h->gen, &n, (char **)&opt);
  if (result != 0) {
    assert(opt != (opt_ptr)NULL);
    *name = opt->name;
    *value = opt->value;
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Frees an option generator for an option handler.]

  Description        [After generating all items in a generation
  sequence, this routine must be called to reclaim the resources
  associated with the created generator.] 

  SideEffects        [None]

  SeeAlso            [Opts_Gen Opts_GenInit]

******************************************************************************/
void Opts_GenFree(Opts_HandlerPtr h) {
  assert((h != OPTS_ERROR) || (h != (Opts_HandlerPtr)NULL));
  assert(h->gen != (st_generator *)NULL);
  
  st_free_gen(h->gen);
  h->gen = (st_generator *)NULL;
}

/**Function********************************************************************

  Synopsis           [Prints all the options on a file]

  Description        [Prints all the options stored in the option
  handler on a given file.]

  SeeAlso            [Opts_GenFree Opts_Gen Opts_GenInit Opts_PrintAllOptions]

******************************************************************************/
void Opts_PrintAllOptions(Opts_HandlerPtr h, FILE * fd) {
  char * name;
  char * value;

  assert(fd != (FILE *)NULL);

  Opts_ForeachOption(h, &name, &value) {
    fprintf(fd, "\t%s = %s\n", name, value);
  }
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/
static char * opts_strsav(const char *s)
{
    return(strcpy(ALLOC(char, strlen(s)+1), s));
}

static boolean opts_isinteger(const char *name) {
  int l;
  char * e[1];
  boolean result = false;

  e[0] = "";
  l = (int)strtol(name, e, 10);
  if (strcmp(e[0], "") == 0) {
    result = true;
  }
  return(result);
}

static void * opts_getinteger(const char *value) {
  int result;
  char * e[1];

  e[0] = "";
  result = (int)strtol(value, e, 10);
  if (strcmp(e[0], "") != 0) {
    result = (int)OPTS_VALUE_ERROR;
  }
  return((void *)result);
}

static opt_ptr option_alloc(void)
{
  opt_ptr result;

  result = ALLOC(opt_rec, 1);
  if (result == NIL(opt_rec)) {
    fprintf(nusmv_stderr, "option_alloc: unable to allocate option entry.\n");
    return((opt_ptr)NULL);
  }
  result->name            = (char *)NULL;
  result->default_value   = (char *)NULL;
  result->value           = (char *)NULL;
  result->possible_values = ovl_create_empty();
  result->check           = (Opts_CheckFnType)NULL;
  result->getvalue        = (Opts_ReturnFnType)NULL;
  return(result);
}

static void option_free(opt_ptr * p) {
  opt_ptr o = *p;
  assert( o != (opt_ptr)NULL);

  if (o->name != (char *)NULL) FREE(o->name);
  if (o->default_value != (char *)NULL) FREE(o->default_value);
  if (o->value != (char *)NULL) FREE(o->value);
  if (ovl_isnotempty(o->possible_values) == 1) {
    ovl_ptr q = o->possible_values;
    ovl_free(&q);
  }
  FREE(*p);
}

static opts_st_retval opts_hash_free(char *key, char *data, char * arg) {
  opt_ptr entry = (opt_ptr)data;

  if (entry != (opt_ptr)NULL) {
    option_free(&entry);
  }
  return(OPTS_DELETE_ENTRY);
}

static opt_ptr option_create(
 const char * name,
 const char * default_value,
 const char * value,
 ovl_ptr       pvalues,
 Opts_CheckFnType check,
 Opts_ReturnFnType getvalue)
{
  opt_ptr result = option_alloc();

  if (result != (opt_ptr)NULL) {
    result->name            = opts_strsav(name);
    result->default_value   = opts_strsav(default_value);
    result->value           = opts_strsav(value);
    result->possible_values = pvalues;
    result->check           = check;
    result->getvalue        = getvalue;
  }
  return(result);
}

static boolean opt_enum_check(const char * value, ovl_ptr l) {
  boolean result = false;
  if (ovl_ispresent(l, value) == 1) {
    result = true;
  }
  return(result);
}

static void * opt_enum_get(const char * value, ovl_ptr l) {
  void * result = OPTS_VALUE_ERROR;
  int found = 0;

  for ( ; (ovl_isnotempty(l) && (found == 0)); l = ovl_get_next(l)) {
    if (strcmp(l->values, value) == 0) {
      found = 1;
      result = (void *)l->valuee;
    }
  }
  return(result);
}

/*
static void * opt_bool_get(const char * value) {
  return((void *)(strcmp(value, OPTS_FALSE_VALUE) ? false : true));
}
*/

static ovl_ptr ovl_rec_alloc(void) {
  ovl_ptr result;

  result = ALLOC(ovl_rec, 1);
  if (result == (ovl_ptr)NULL) {
    fprintf(nusmv_stderr, "ovl_rec_alloc: unable to allocate a value record.\n");
  }
  else {
    result->next = ovl_create_empty();
    result->values = (char *)NULL;
    result->valuee = (int)OPTS_VALUE_ERROR;
  }
  return(result);
}

static ovl_ptr ovl_create_empty(void) {
  return((ovl_ptr)NULL);
}

static int ovl_isempty(ovl_ptr l) {
  return((l == ovl_create_empty()));
}

static int ovl_isnotempty(ovl_ptr l) {
  return((l != ovl_create_empty()));
}

static ovl_ptr ovl_create(const char * values, int valuee) {
  ovl_ptr result = (ovl_ptr)NULL;

  result = ovl_rec_alloc();
  if (result != (ovl_ptr)NULL) {
    result->values = opts_strsav(values);
    result->valuee = valuee;
  }
  return(result);
}

static char * ovl_get_values(ovl_ptr l) {
  assert(ovl_isempty(l) != 1);
  return(l->values);
}

static int ovl_get_valuee(ovl_ptr l) {
  assert(ovl_isempty(l) != 1);
  return(l->valuee);
}

static ovl_ptr ovl_get_next(ovl_ptr l) {
  assert(ovl_isnotempty(l) == 1);
  return(l->next);
}

static ovl_ptr ovl_set_next(ovl_ptr l, ovl_ptr n) {
  assert(ovl_isnotempty(l) == 1);
  l->next = n;
  return(l);
}

static void ovl_free(ovl_ptr *l) {
  ovl_ptr p;
  
  p = *l;
  while (ovl_isnotempty(p)) {
    ovl_ptr q = p;

    p = ovl_get_next(p);
    FREE(q->values);
    FREE(q);
  }
}

static int ovl_ispresent(ovl_ptr l, const char *value) {
  int result = 0;

  for ( ; (ovl_isnotempty(l) && (result == 0)); l = ovl_get_next(l)) {
    result = (strcmp(l->values, value) == 0);
  }
  return(result);
}

static int print_opl(ovl_ptr l) {
  for ( ; (ovl_isnotempty(l) == 1); l = ovl_get_next(l)) {
    printf("Value = %s\t%d\n", l->values, l->valuee); 
  }
  return(0);
}

static int check_boolean(ovl_ptr l) {
  int result = 0;
 
  for ( ; (ovl_isnotempty(l) && (result == 0)); l = ovl_get_next(l)) {
    result = ((strcmp(l->values, OPTS_TRUE_VALUE) == 0) || 
              (strcmp(l->values, OPTS_FALSE_VALUE) == 0));
  }
  return(result);
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/



