/**CHeaderFile*****************************************************************

  FileName    [opts.h]

  PackageName [opts]

  Synopsis    [Generic handler of options]

  Description [Generic handler of options. An option is uniquely
  identified in an option handler by:
  <ul>
    <li> a name (it is the string of its name). 
    <li> a default value (it is the string of its name). 
    <li> a value (it is the string of its name). 
  </ul>
  When registering an option the user must specify two functions. The
  first is responsible of checking that the value passed while setting
  a value is a valid value for the given option. The second function
  is a function that transforms the stroed value in a value suitable
  to be used in the calling program. <br>

  For boolean options are provided special methods to register the
  option and for setting and getting a value associated to it.<br>

  For enumerative options are provided special methods to register the
  option and for setting and getting a value associated to it. An
  enumertive option is registered by providing an array of structures
  of type Opts_EnumRec. Similarly to the below declaration:<br>
  <pre>
    typedef enum {foo1, ...., fooN} fooenumtype;
    Opts_EnumRec foo[] = {"foo1", foo1, 
                          "foo2", foo2, 
                          ....
                          "fooN", fooN};
   ....
   handler = Opts_HandlerAlloc();
   Opts_HandlerRegisterEnumOption(handler, "foooption", "foo1", foo, N);

  if (Opts_HandlerGetEnumOptionValue(handler, "foooption") == foo2) {
     ...
  }

  ...

  switch(Opts_HandlerGetEnumOptionValue(handler, "foooption")) {
    case foo1:
    ...
    case fooN:
    ...
    default:
    ...
  }
  </pre>
  ]

  Author      [Marco Roveri]

  Copyright   [ 
  This file is part of the ``simulator'' package of NuSMV version 2.
  Copyright (C) 2003 by ITC-irst.

  NuSMV version 2 is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>.
  ]

  Revision    [$Id: opts.h,v 1.3 2003/10/09 10:07:54 nusmv Exp $]

******************************************************************************/

#ifndef _opts_h
#define _opts_h

#include <stdlib.h>
#include <stdio.h>
#include "util.h"
#include "utils.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

typedef boolean (*Opts_CheckFnType)(const char *);
typedef void * (*Opts_ReturnFnType)(const char *);
typedef struct _Opts_HandlerRec   Opts_HandlerRec;
typedef struct _Opts_HandlerRec * Opts_HandlerPtr;
typedef struct _Opts_EnumRec Opts_EnumRec;

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/
struct _Opts_EnumRec {
  char *v;
  int  e;
};

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/
#define OPTS_ERROR (Opts_HandlerPtr)-1
#define OPTS_TRUE_VALUE "1"
#define OPTS_FALSE_VALUE "0"

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/
#define OPTS_ERROR (Opts_HandlerPtr)-1
#define OPTS_VALUE_ERROR (void *)-9999
#define OPTS_TRUE_VALUE  "1"
#define OPTS_FALSE_VALUE "0"


/**Macro***********************************************************************

  Synopsis     [Operates on each entry of the option handler]

  Description  [Operates on each entry of the option handler. name and
  value must be declared to be char **.]

  SeeAlso      [Opts_GenInit Opts_Gen Opts_GenFree]

******************************************************************************/
#define Opts_ForeachOption(h, name, value) \
    for (Opts_GenInit(h); Opts_Gen(h, name, value) || (Opts_GenFree(h),0); )


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/
EXTERN Opts_HandlerPtr Opts_HandlerAlloc ARGS((void)); 
EXTERN void Opts_HandlerFree ARGS((Opts_HandlerPtr * h)); 
EXTERN boolean Opts_HandlerIsOptionRegistered ARGS((Opts_HandlerPtr h, const char * name)); 
EXTERN boolean Opts_HandlerIsOptionNotRegistered ARGS(( Opts_HandlerPtr h, const char * name)); 
EXTERN boolean Opts_HandlerRegisterOption ARGS((Opts_HandlerPtr h, const char * name, const char * def, Opts_CheckFnType check, Opts_ReturnFnType get)); 
EXTERN boolean Opts_HandlerRegisterBoolOption ARGS(( Opts_HandlerPtr h, const char * name, boolean value));
EXTERN boolean Opts_HandlerRegisterEnumOption ARGS((Opts_HandlerPtr h, const char * name, const char * def, Opts_EnumRec pv[], int npv));
EXTERN boolean Opts_HandlerRegisterIntOption ARGS(( Opts_HandlerPtr h, const char * name, int value));

EXTERN boolean Opts_HandlerUnregisterOption ARGS((Opts_HandlerPtr h, const char * name));
EXTERN boolean Opts_HandlerSetOptionValue ARGS((Opts_HandlerPtr h, const char * name, const char * value)); 
EXTERN boolean Opts_HandlerSetEnumOptionValue ARGS((Opts_HandlerPtr h, const char * name, const char * value)); 
EXTERN boolean Opts_HandlerSetBoolOptionValue ARGS((Opts_HandlerPtr h, const char * name, boolean value)); 
EXTERN boolean Opts_HandlerResetOptionValue ARGS((Opts_HandlerPtr h, const char * name)); 
EXTERN void * Opts_HandlerGetOptionValue ARGS((Opts_HandlerPtr h, const char * name)); 
EXTERN int Opts_HandlerGetEnumOptionValue ARGS((Opts_HandlerPtr h, const char * name)); 
EXTERN boolean Opts_HandlerGetBoolOptionValue ARGS((Opts_HandlerPtr h, const char * name)); 
EXTERN int Opts_HandlerGetIntOptionValue ARGS((Opts_HandlerPtr h, const char * name));

EXTERN void Opts_GenInit ARGS((Opts_HandlerPtr h)); 
EXTERN int Opts_Gen ARGS((Opts_HandlerPtr h, char **name, char **value)); 
EXTERN void Opts_GenFree ARGS((Opts_HandlerPtr h)); 
EXTERN void Opts_PrintAllOptions ARGS((Opts_HandlerPtr h, FILE * fd));

/**AutomaticEnd***************************************************************/

#endif /* _ */
