/**CFile***********************************************************************

  FileName    [dsimulate.c]

  PackageName [simulator]

  Synopsis    [The new simulation engine]

  Description [This file provides the first version of the new
  interactive symulator.]

  SeeAlso     [optional]

  Author      [Marco Roveri]

  Copyright   [This file is part of the ``simulator'' package of NuSMV
  version 2.  Copyright (C) 2003 ITC-irst.

  NuSMV version 2 is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2 of
  the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it> or
  email to <nusmv-users@irst.itc.it>.  Please report bugs to
  <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to
  <nusmv@irst.itc.it>.]

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>

#include "util.h"
#include "utils.h"
#include "ustring.h"
#include "node.h"
#include "grammar.h"
#include "dd.h"
#include "bddfsm.h"
#include "compile.h"
#include "dsimulate.h"
#include "opts.h"

static char rcsid[] UTIL_UNUSED = "$Id: dsimulate.c,v 1.6 2003/10/09 10:07:54 nusmv Exp $";

#define NOTIMPLEMENTED \
  fprintf(nusmv_stderr, "Function %s is not implemented.\n", \
          __PRETTY_FUNCTION__), \
  abort()

#define _NO_NAME_ Nil

typedef enum {_SIM_RND_, _SIM_FIX_, _SIM_INT_ } SimulationSelectionMode;
typedef enum {_SIM_USER_, _SIM_GOD_} SimulationMode;
typedef enum {_SIM_TAB_VAR_FIXED_, _SIM_TAB_VAR_UNBOUND_} SimulationTabVarMode;

typedef struct _sim_opt_struct Simulation_OptionsStructure;
typedef struct _sim_opt_struct * Simulation_OptionsPtr;

struct _sim_opt_struct {
  int steps;
  SimulationSelectionMode selection_mode;
  SimulationMode mode;
  SimulationTabVarMode tabmode;
};

/* used to communicate that the simulation stopped among the different levels */
/* it could be embedded in the opt structure passed to low level simulation routines. */
static boolean breaked_simulation = false;

extern DdManager * dd_manager;
extern Opts_HandlerPtr simopthr;
extern node_ptr all_symbols;
extern node_ptr proc_selector_internal_vname;

static void print_vlist ARGS((FILE * fd, node_ptr l, node_ptr name));
static node_ptr extract_values ARGS((node_ptr name, BddFsm fsm, bdd_ptr set));
static node_ptr get_element ARGS((node_ptr l, int k));
static node_ptr mk_equal ARGS((node_ptr l, node_ptr r));
static node_ptr mk_true ARGS((void));
static node_ptr mk_false ARGS((void));
static void simulate_print_path_god ARGS((FILE * fd, BddFsm fsm, node_ptr path));
static void simulate_save_path_god ARGS((FILE * fd, BddFsm fsm, node_ptr path));

static void simulatePrintState ARGS((FILE * fd, BddFsm fsm, bdd_ptr s));
static void simulatePrintInput ARGS((FILE * fd, BddFsm fsm, bdd_ptr a));
static void free_simulation_path ARGS((node_ptr path));
static void simulate_print_path_generic ARGS((FILE *fd, BddFsm fsm, node_ptr path));
static void simulate_save_path_generic ARGS((FILE *fd, BddFsm fsm, node_ptr path));

static int simulate_int_print_trace ARGS((void));
static int simulate_int_save_trace ARGS((void));

static int check_interactive_read ARGS((int r, int k, int n));
static int is_to_avoid_name ARGS((node_ptr name));
static int ask_user_yes_no ARGS((const char * msg));
static boolean simulate_continue_simulation ARGS((void));
static void simulate_print_state ARGS((FILE *fd, node_ptr state));
static void simulate_print_input ARGS((FILE *fd, node_ptr state));
static void print_trace_state ARGS((bdd_ptr s, node_ptr symbols, int changes_only));


/* Dsimulation option structure */
Simulation_OptionsPtr simulateOptAlloc() {
  Simulation_OptionsPtr result;

  result = ALLOC(Simulation_OptionsStructure, 1);
  assert(result != (Simulation_OptionsPtr)NULL);
  return(result);
}

void simulateOptSetSelectionMode(Simulation_OptionsPtr p, SimulationSelectionMode m) {
  assert(p != (Simulation_OptionsPtr)NULL);
  p->selection_mode = m;
}

SimulationSelectionMode simulateOptGetSelectionMode(Simulation_OptionsPtr p) {
  assert(p != (Simulation_OptionsPtr)NULL);
  return(p->selection_mode);
}

void simulateOptSetMode(Simulation_OptionsPtr p, SimulationMode m) {
  assert(p != (Simulation_OptionsPtr)NULL);
  p->mode = m;
}

SimulationMode simulateOptGetMode(Simulation_OptionsPtr p) {
  assert(p != (Simulation_OptionsPtr)NULL);
  return(p->mode);
}

void simulateOptSetTabVarMode(Simulation_OptionsPtr p, SimulationTabVarMode m) {
  assert(p != (Simulation_OptionsPtr)NULL);
  p->tabmode = m;
}

SimulationTabVarMode simulateOptGetTabVarMode(Simulation_OptionsPtr p) {
  assert(p != (Simulation_OptionsPtr)NULL);
  return(p->tabmode);
}

void simulateSetOptMaxSteps(Simulation_OptionsPtr p, int s) {
  assert(p != (Simulation_OptionsPtr)NULL);
  p->steps = s;
}

int simulateGetOptMaxSteps(Simulation_OptionsPtr p) {
  assert(p != (Simulation_OptionsPtr)NULL);
  return(p->steps);
}

/* option handling */
void DSimulate_RegisterOptions(void) {
  Opts_EnumRec o1[] = {{"user", _SIM_USER_}, {"god", _SIM_GOD_}};
  Opts_EnumRec o2[] = {{"rnd", _SIM_RND_}, {"fix", _SIM_FIX_}, {"int", _SIM_INT_}};
  Opts_EnumRec o3[] = {{"fix", _SIM_TAB_VAR_FIXED_}, {"nofix", _SIM_TAB_VAR_UNBOUND_}};

  if (Opts_HandlerRegisterEnumOption(simopthr,
                                     "simulate_mode",
                                     "god",
                                     o1,
                                     sizeof(o1)/sizeof(Opts_EnumRec)) != true) {
    fprintf(nusmv_stderr, "DSimulate_RegisterOptions: Unable to register option \"%s\"\n", "simulate_mode");
    nusmv_exit(1);
  }
  if (Opts_HandlerRegisterEnumOption(simopthr,
                                     "simulate_state_sel_mode",
                                     "rnd",
                                     o2,
                                     sizeof(o2)/sizeof(Opts_EnumRec)) != true) {
    fprintf(nusmv_stderr, "DSimulate_RegisterOptions: Unable to register option \"%s\"\n", "simulate_state_sel_mode");
    nusmv_exit(1);
  }

  if (Opts_HandlerRegisterEnumOption(simopthr,
                                     "simulate_tab_vars_mode",
                                     "nofix",
                                     o3,
                                     sizeof(o3)/sizeof(Opts_EnumRec)) != true) {
    fprintf(nusmv_stderr, "DSimulate_RegisterOptions: Unable to register option \"%s\"\n", "simulate_tab_vars_mode");
    nusmv_exit(1);
  }

  if (Opts_HandlerRegisterIntOption(simopthr,
                                    "simulate_maxsteps",
                                    10) != true) {
    fprintf(nusmv_stderr, "DSimulate_RegisterOptions: Unable to register option \"%s\"\n", "simulate_maxsteps");
    nusmv_exit(1);
  }

  if (Opts_HandlerRegisterBoolOption(simopthr,
                                     "simulate_active",
                                    false) != true) {
    fprintf(nusmv_stderr, "DSimulate_RegisterOptions: Unable to register option \"%s\"\n", "simulate_active");
    nusmv_exit(1);
  }
}

void DSimulate_UnRegisterOptions(void) {
  Opts_HandlerUnregisterOption(simopthr, "simulate_maxsteps");
  Opts_HandlerUnregisterOption(simopthr, "simulate_state_sel_mode");
  Opts_HandlerUnregisterOption(simopthr, "simulate_tab_vars_mode");
  Opts_HandlerUnregisterOption(simopthr, "simulate_mode");
  Opts_HandlerUnregisterOption(simopthr, "simulate_active");
}

int DSimulate_ParseCLOptions(char *** argvv, int * argcc) {
  char ** argv = *argvv;
  int argc = *argcc;
  int result = 0;

  /* domain simulation options */
  if (strcmp(*argv,"-simulate") == 0) {
    argv++; argc --;
    Opts_HandlerSetBoolOptionValue(simopthr, "simulate_active", true);
    result = 1;
  }
  else if (strcmp(*argv,"-simulate_mode") == 0) {
    if (argc < 2) {
      fprintf(nusmv_stderr, "The \"-simulate_mode\" command line option requires an argument (user|god).\n");
      nusmv_exit(1);
    }
    argv++; argc -= 2;
    if (strcmp(*argv, "user") == 0) {
      Opts_HandlerSetEnumOptionValue(simopthr, "simulate_mode", "user");
    }
    else if (strcmp(*argv, "god") == 0) {
      Opts_HandlerSetEnumOptionValue(simopthr, "simulate_mode", "god");
    }
    else {
      fprintf(nusmv_stderr, "The \"-simulate_mode\" command line option requires an argument (user|god).\n");
      nusmv_exit(1);
    }
    argv++;
    result = 1;
  }
  else if (strcmp(*argv,"-simulate_tab_vars_mode") == 0) {
    if (argc < 2) {
      fprintf(nusmv_stderr, "The \"-simulate_tab_vars_mode\" command line option requires an argument (fix|nofix).\n");
      nusmv_exit(1);
    }
    argv++; argc -= 2;
    if (strcmp(*argv, "fix") == 0) {
      Opts_HandlerSetEnumOptionValue(simopthr, "simulate_tab_vars_mode", "fix");
    }
    else if (strcmp(*argv, "nofix") == 0) {
      Opts_HandlerSetEnumOptionValue(simopthr, "simulate_tab_vars_mode", "nofix");
    }
    else {
      fprintf(nusmv_stderr, "The \"-simulate_tab_vars_mode\" command line option requires an argument (fix|nofix).\n");
      nusmv_exit(1);
    }
    argv++;
    result = 1;
  }
  else if (strcmp(*argv,"-simulate_sel_mode") == 0) {
    if (argc < 2) {
      fprintf(nusmv_stderr, "The \"-simulate_sel_mode\" command line option requires an argument (rnd|fix|int).\n");
      nusmv_exit(1);
    }
    argv++; argc -= 2;
    if (strcmp(*argv, "rnd") == 0) {
      Opts_HandlerSetEnumOptionValue(simopthr, "simulate_state_sel_mode", "rnd");
    }
    else if (strcmp(*argv, "fix") == 0) {
      Opts_HandlerSetEnumOptionValue(simopthr, "simulate_state_sel_mode", "fix");
    }
    else if (strcmp(*argv, "int") == 0) {
      Opts_HandlerSetEnumOptionValue(simopthr, "simulate_state_sel_mode", "int");
    }
    else {
      fprintf(nusmv_stderr, "The \"-simulate_sel_mode\" command line option requires an argument (rnd|fix|int).\n");
      nusmv_exit(1);
    }
    argv++;
    result = 1;
  }
  else if (strcmp(*argv,"-simulate_maxsteps") == 0) {
    if (argc < 2) {
      fprintf(nusmv_stderr, "The \"dsimulate_maxsteps\" command line option requires an argument #natural.\n");
      nusmv_exit(1);
    }
    argv++; argc -= 2;
    /* trick here, instead of calling the integer method */
    if (Opts_HandlerSetEnumOptionValue(simopthr,
                                       "simulate_maxsteps",
                                       (const char *)*argv) == false) {
      fprintf(nusmv_stderr, "The \"dsimulate_maxsteps\" command line option requires an argument #natural.\n");
      nusmv_exit(1);
    }
    argv++;
    result = 1;
  }
  *argvv = argv;
  *argcc = argc;
  return(result);
}

void DSimulate_PrintCLOptions(FILE * fd) {
  fprintf(fd, "\t===================  Domain Simulation Options  ====================\n\n");
  /* domain */
  fprintf(fd, "\t -simulate\t\t\tActivates the domain simulation mode.\n");
  fprintf(fd, "\t -simulate_mode <mode>\t\tSpecifies simulation mode:\n\t\t\t\t\t <mode>={god|user}.\n");
  fprintf(fd, "\t -simulate_sel_mode <mode>\tSpecifies simulation state selection\n\t\t\t\t\t mode: <mode>={rnd|fix|int}.\n");
  fprintf(fd, "\t -simulate_tab_vars_mode <mode>\t\tSpecifies the selection mode for tableau variables:\n\t\t\t\t\t <mode>={fix|nofix}.\n");
  fprintf(fd, "\t -simulate_maxsteps #int\tSpecifies maximum number of simulation\n\t\t\t\t\t steps.\n");
  fprintf(fd, "\n");
  fprintf(fd, "\n\t======================================================================\n\n");
}
 

/* ********************

       STATE SELECTION

******************** */

/* as default we select the first element */
node_ptr select_state_fix(node_ptr name, node_ptr vals) {
  node_ptr result;
  
  assert(vals != Nil);
  result = car(vals);
  return(result);
}

/* random selection of a value among the possible ones */
node_ptr select_state_rnd(node_ptr name, node_ptr vals) {
  node_ptr result;
  int l = llength(vals);
  
  assert(vals != Nil);

  if (l == 1) {
    result = car(vals);
  }
  else {
    int j = (random() % l);

    result = get_element(vals, j);
  }
  return(result);
}

node_ptr select_state_interactive(node_ptr name, node_ptr vals) {
  node_ptr result;
  int k,r;
  int n =  llength(vals);
  
  assert(vals != Nil);
  
  /* if only one element, then it is automatically chosen */
  if (n == 1) return(car(vals));
  fprintf(nusmv_stdout, "State variable \n\t\"");
  print_node(nusmv_stdout, name);
  fprintf(nusmv_stdout, "\"\ncan assume one of the following values:\n\n");
  print_vlist(nusmv_stdout, vals, name);

  do {
    fprintf(stdout, "\nChoose a number among %d--%d (-1 to stop): ",0,n-1);
    r = scanf("%d", &k);
    getchar();
    fflush(stdin);
    if (k == -1) {
      k = 0;
      r = EOF;
    }
  } while (check_interactive_read(r, k, n));
  /* as default we choose the first element */
  if (r == EOF) {
    k = 0;
    breaked_simulation = true;
  }
  result = get_element(vals, k);
  return(result);
}

/* 
   Selects one state in "states".

   Actually this function returns a BDD. An optimization could be to
   return a pair (Bdd . ass) where ass is the sexp corresponding to
   the BDD. This would avoid call to BddFsm_WffCond in the plan
   simulation. */
bdd_ptr select_state(BddFsm fsm, bdd_ptr states, Simulation_OptionsPtr opt) {
  bdd_ptr sstates = bdd_dup(states);
  node_ptr vars = BddFsm_GetStateVars(fsm);
  bdd_ptr result = bdd_one(dd_manager);

  for (  ; vars != Nil; vars = cdr(vars)) {
    node_ptr value;
    node_ptr name = car(vars);
    node_ptr pvalue = extract_values(name, fsm, sstates);
    int avoid_name = is_to_avoid_name(name);

    assert(pvalue != Nil);
    /* skip of the "_constant" dummy variable by assigning it a
       default value */
    
    if ((avoid_name == 0) || ((avoid_name == 1) &&
                              (simulateOptGetTabVarMode(opt) ==
                               _SIM_TAB_VAR_FIXED_))) {
      if (avoid_name == 1) {
        value = select_state_rnd(name, pvalue);
      }
      else { 
        if (breaked_simulation == true) {
          value = select_state_fix(name, pvalue);
        }
        else {
          switch(simulateOptGetSelectionMode(opt)) {
          case _SIM_RND_:
            value = select_state_rnd(name, pvalue);
            break;
          case _SIM_FIX_:
            value = select_state_fix(name, pvalue);
            break;
          case _SIM_INT_:
            value = select_state_interactive(name, pvalue);
            break;
          default:
            fprintf(nusmv_stderr, "select_state: unknwon selection mode: %d\n",
                    simulateOptGetSelectionMode(opt));
            value = Nil;
            nusmv_exit(1);
            break;
          }
        }
      }
      {
        bdd_ptr v = BddFsm_EvalCond(fsm, mk_equal(name, value));

        bdd_and_accumulate(dd_manager, &sstates, v);
        bdd_and_accumulate(dd_manager, &result, v);
        bdd_free(dd_manager, v);
      }
    }
  }
  fprintf(nusmv_stdout, "State is: ");
  simulatePrintState(nusmv_stdout, fsm, result);
  fprintf(nusmv_stdout, "\n");
  /* theorically the following assertion must hold the input variables
     possibly in states could cause teh failure of this assertion */
  assert(sstates == result);
  bdd_free(dd_manager, sstates);
  return(result);
}



/* ********************

       INPUT SELECTION

******************** */


/* as default we select the first element */
node_ptr select_input_fix(node_ptr name, node_ptr vals) {
  node_ptr result;
  
  assert(vals != Nil);
  result = car(vals);
  return(result);
}

/* random selection of a value among the possible ones */
node_ptr select_input_rnd(node_ptr name, node_ptr vals) {
  node_ptr result;
  int l = llength(vals);
  
  assert(vals != Nil);

  if (l == 1) {
    result = car(vals);
  }
  else {
    int j = (random() % l);

    result = get_element(vals, j);
  }
  return(result);
}

node_ptr select_input_interactive(node_ptr name, node_ptr vals) {
  node_ptr result;
  int k,r;
  int n = llength(vals);
  
  assert(vals != Nil);
  
  /* if only one element, then it is automatically chosen */
  if (n == 1) return(car(vals));
  fprintf(nusmv_stdout, "Input variable\n\t \"");
  print_node(nusmv_stdout, name);
  fprintf(nusmv_stdout, "\"\ncan assume one of the following values:\n\n");
  print_vlist(nusmv_stdout, vals, _NO_NAME_);

  do {
    fprintf(stdout, "\nChoose a number among %d--%d (-1 to stop): ",0,n-1);
    r = scanf("%d", &k);
    getchar();
    fflush(stdin);
    if (k == -1) {
      k = 0;
      r = EOF;
    }
  } while (check_interactive_read(r, k, n));
  /* as default we choose the first element */
  if (r == EOF) {
    k = 0;
    breaked_simulation = true;
  }
  result = get_element(vals, k);
  return(result);
}

/* 
   Selects one input assignment among the inputs applicable in
   "state" according to the input applicability relation "sa".

   Actually this function returns a BDD. An optimization could be to
   return a pair (Bdd . ass) where ass is the sexp corresponding to
   the BDD. This would avoid call to BddFsm_WffCond in the plan
   simulation. */
bdd_ptr select_input(BddFsm fsm, bdd_ptr state, bdd_ptr sa,
                      Simulation_OptionsPtr opt)
{
  node_ptr vars;
  bdd_ptr applicable_inputs;
  node_ptr inputs = BddFsm_GetInputVars(fsm);
  bdd_ptr result = bdd_one(dd_manager);
  bdd_ptr cube = BddFsm_GetStateSupport(fsm);

  /* the set of possible applicable inputs */
  applicable_inputs = bdd_and_abstract(dd_manager, sa, state, cube);

  if (bdd_is_zero(dd_manager, applicable_inputs)) {
    fprintf(nusmv_stderr, "applicable_inputs is zero\n");
    dd_printminterm(dd_manager, state);
    fprintf(nusmv_stderr, "kkkkkkkkkkkkkk\n");
    dd_printminterm(dd_manager, cube);
  }
  for (vars = inputs; vars != Nil; vars = cdr(vars)) {
    node_ptr value;
    node_ptr name = car(vars);
    node_ptr pvalue = extract_values(name, fsm, applicable_inputs);

    assert(pvalue != Nil);

    if (breaked_simulation == true) {
      value = select_input_fix(name, pvalue);
    }
    else {
      switch(simulateOptGetSelectionMode(opt)) {
      case _SIM_RND_:
        value = select_input_rnd(name, pvalue);
        break;
      case _SIM_FIX_:
        value = select_input_fix(name, pvalue);
        break;
      case _SIM_INT_:
        value = select_input_interactive(name, pvalue);
        break;
      default:
        fprintf(nusmv_stderr, "select_input: unknwon selection mode: %d\n",
                simulateOptGetSelectionMode(opt));
        value = Nil;
        nusmv_exit(1);
        break;
      }
    }
    {
      bdd_ptr v = BddFsm_EvalCond(fsm, mk_equal(name, value));

      bdd_and_accumulate(dd_manager, &applicable_inputs, v);
      bdd_and_accumulate(dd_manager, &result, v);
      bdd_free(dd_manager, v);
    }
  }
  /* put here some printing */
  if (bdd_isnot_zero(dd_manager, result) && 
      bdd_isnot_one(dd_manager, result)) {
    fprintf(nusmv_stdout, "Inputs are: \n");
    simulatePrintInput(nusmv_stdout, fsm, result);
    fprintf(nusmv_stdout, "\n");
  }
  /* This assertion must hold */
  assert(applicable_inputs == result);
  bdd_free(dd_manager, applicable_inputs);
  bdd_free(dd_manager, cube);
  return(result);
}


/* ********************

       DOMAIN SIMULATION 

******************** */
/* comment the below define to avoid handlign with inputs. */
#define HANDLE_INPUTS

/*
  The set of states with a successor is 

   {s : \exists s,a, (s,a,s') \in R}

   While the set of fair states is the set of those states that are
   the origin of a fair path (being a fair path an infinite path along
   with the fairness conditions infinitely often hold). It seems not
   necessary to restrict to states with a successors (i.e. define
   RESTRICT_TO_STATES_WITH_SUCCESSOR whenever the
   RESTRICT_TO_FAIR_STATES is defined. The use of
   RESTRICT_TO_STATES_WITH_SUCCESSOR seems an approximations, but it
   might be the case that we end-up in a dead-end state, since
   guaranteeing to have a successor does not guarantee that we are
   restricting to an infinite path.
*/
#define RESTRICT_TO_FAIR_STATES
#define RESTRICT_TO_STATES_WITH_SUCCESSOR
void domain_simulate(BddFsm fsm, Simulation_OptionsPtr options) 
{
  int step, maxsteps;
  node_ptr path = Nil;
  node_ptr cobs = Nil;
  bdd_ptr current_state;
  bdd_ptr initial_set_of_states;
  bdd_ptr states_input_with_succes;
  bdd_ptr states = BddFsm_GetStates(fsm);
#ifdef RESTRICT_TO_FAIR_STATES
  bdd_ptr fair_states = BddFsm_GetFairStates(fsm);
#else
  bdd_ptr fair_states = bdd_one(dd_manager);
#endif
  bdd_ptr one = bdd_one(dd_manager);
  bdd_ptr current_input = bdd_zero(dd_manager);

  breaked_simulation = false;

  maxsteps = simulateGetOptMaxSteps(options);

#ifdef RESTRICT_TO_STATES_WITH_SUCCESSOR
  /* extracting the set of pairs state.input such that input is
     applicable in state, i.e. the set of states with a successor when
     no input are present in the model */
  {
    bdd_ptr p = bdd_and(dd_manager, fair_states, states);

    states_input_with_succes = BddFsm_WeakImageBwd(fsm, p);
    bdd_free(dd_manager, p);
  }
#else
  states_input_with_succes = bdd_one(dd_manager);
#endif

  /* Extracting the set of initial states */
  initial_set_of_states = BddFsm_GetInit(fsm);

  /* intersect with the set of admissible model states */
  bdd_and_accumulate(dd_manager, &initial_set_of_states, states);

  /* intersect with the set of fair states */
  bdd_and_accumulate(dd_manager, &initial_set_of_states, fair_states);
    
  /* intersect with the set of states that have a successor state */
  bdd_and_accumulate(dd_manager, &initial_set_of_states, states_input_with_succes);
  
  /* Extracting a state from the set of initial states according to
     the simulation options */
  fprintf(nusmv_stdout, "===> Initial State\n");
  current_state = select_state(fsm, initial_set_of_states, options);

  /* it is no longer needed */
  bdd_free(dd_manager, initial_set_of_states);

  if (breaked_simulation == false) {
    /* would we continue with the simulation? */
    breaked_simulation = simulate_continue_simulation();
  }

  /* transform this in do while which appear to be more compact */
  for (step = 0; ((step < maxsteps) && (breaked_simulation == false)); step++) {
    bdd_ptr image;
    bdd_ptr csa;

    fprintf(nusmv_stdout, "===> Simulation Step %d\n", step);

    if (bdd_is_zero(dd_manager, current_state)) {
      fprintf(nusmv_stderr, "The current state is zero.\n");
    }

    /* selection of an input among the applicable inputs in the
       chosen state */
    bdd_free(dd_manager, current_input);
#ifdef HANDLE_INPUTS
    current_input = select_input(fsm, current_state,
                                   states_input_with_succes, options);

    if (bdd_is_zero(dd_manager, current_input) == 1) {
      fprintf(nusmv_stderr, "WARNING: no applicable inputs in the chosen state.\n");
      breaked_simulation = true;
    }
#else
    current_input = bdd_one(dd_manager);
#endif

    /* Initializing the list of BDDs */
    path = cons(cons(cons((node_ptr)bdd_dup(current_state),
                          cobs),
                     (node_ptr)bdd_dup(current_input)), path);

    /* building the current pair state.input to constrain image fwd */
    csa = bdd_and(dd_manager, current_input, current_state);

    /* computing image forward */
    image = BddFsm_ImageFwd(fsm, csa);

    /* csa no longer needed */
    bdd_free(dd_manager, csa);

    if (bdd_is_zero(dd_manager, image)) {
      fprintf(nusmv_stderr, "domain_simulate: image is empty\n");
    }
    /* intersecting with the set of admissible model states */
    bdd_and_accumulate(dd_manager, &image, states);

    if (bdd_is_zero(dd_manager, image)) {
      fprintf(nusmv_stderr, "domain_simulate: image restricted to invariant states is empty\n");
    }
    /* intersect with the set of fair states */
    bdd_and_accumulate(dd_manager, &image, fair_states);

    if (bdd_is_zero(dd_manager, image)) {
      fprintf(nusmv_stderr, "domain_simulate: image restricted to fair states is empty\n");
    }
    /* intersect with the set of states that have a successor state */
    bdd_and_accumulate(dd_manager, &image, states_input_with_succes);

    bdd_free(dd_manager, current_state);
    if (bdd_is_zero(dd_manager, image)) {
      /* The current states is the empty set we reached a deadend */
      fprintf(nusmv_stderr, "The current state has no successor states.\n");
      fprintf(nusmv_stderr, "Stopping the simulation!!!!!\n");
      current_state = bdd_zero(dd_manager);
      breaked_simulation = true;
    }
    else {
      /* The current states is the new one */
      current_state = select_state(fsm, image, options);
    }
    /* image no longer needed */
    bdd_free(dd_manager, image);

    if (breaked_simulation == false) {
      /* would we continue with the simulation? */
      breaked_simulation = simulate_continue_simulation();
    }
  }

  /* The last state of the path is the one resulting from the exit of
     the loop. */
  path = cons(cons(cons((node_ptr)bdd_dup(current_state),
                        cobs),
                   (node_ptr)bdd_zero(dd_manager)), path);

  /* the path needs to be reversed */
  path = reverse(path);

  simulate_print_path_god(nusmv_stdout, fsm, path);

  simulate_save_path_god(nusmv_stdout, fsm, path);

  /* freeing BDDs */
  free_simulation_path(path);
  bdd_free(dd_manager, current_state);
  bdd_free(dd_manager, states);
  bdd_free(dd_manager, current_input);
  bdd_free(dd_manager, one);
}

/* Domain Simulation */
void DSimulate_DomainSimulate(BddFsm fsm) 
{
  Simulation_OptionsPtr opt = simulateOptAlloc();

#ifdef _SIMULATE_PRINT_OPTIONS
  fprintf(nusmv_stdout, "====================================================================\n");
  fprintf(nusmv_stdout, "=======================  Simulation Options  =======================\n");
  fprintf(nusmv_stdout, "====================================================================\n");
  Opts_PrintAllOptions(simopthr, nusmv_stdout);
#endif
  fprintf(nusmv_stdout, "====================================================================\n");
  fprintf(nusmv_stdout, "==================  Starting  Domain  Simulation  ==================\n");
  fprintf(nusmv_stdout, "====================================================================\n");

  simulateOptSetSelectionMode(opt, Opts_HandlerGetEnumOptionValue(simopthr, "simulate_state_sel_mode"));
  simulateOptSetMode(opt, Opts_HandlerGetEnumOptionValue(simopthr, "simulate_mode"));
  simulateSetOptMaxSteps(opt, Opts_HandlerGetIntOptionValue(simopthr, "simulate_maxsteps"));
  simulateOptSetTabVarMode(opt, Opts_HandlerGetEnumOptionValue(simopthr, "simulate_tab_vars_mode"));

  switch(simulateOptGetMode(opt)) {
  case _SIM_GOD_:
    domain_simulate(fsm, opt);
    break;
  default:
    fprintf(nusmv_stderr, "DSimulate_DomainSimulate: Unknown simulation mode.\n");
    nusmv_exit(1);
    break;
  }
  FREE(opt);
  fprintf(nusmv_stdout, "====================================================================\n");
  fflush(nusmv_stdout);
  fflush(nusmv_stderr);
}


void print_tropos_trace(int trace, int verbose) 
{
  int i = 1;
  node_ptr state = Nil;
  node_ptr label = trace_mk_statelabel(trace, i);
  node_ptr symbols = all_symbols;

  clear_print_hash();
  state = lookup_traces_hash(label);
  fprintf(nusmv_stdout,
          "################### Trace number: %d ###################\n", trace);
  while(state != Nil) {
    fprintf(nusmv_stdout, "-> State ");
    print_node(nusmv_stdout, label);
    fprintf(nusmv_stdout, " <-\n");
    set_indent_size(2);
    print_trace_state((bdd_ptr)state, symbols, verbose);
    reset_indent_size();      
    ++i;
    label = trace_mk_statelabel(trace, i);
    state = lookup_traces_hash(label);
  }
}


/* static functions */
static void print_vlist(FILE * fd, node_ptr l, node_ptr name) 
{
  int k;
  for (k=0; l != Nil; l = cdr(l)) {
    fprintf(fd, "\t [%3d] ===>> ",k++);
    if (name == _NO_NAME_) {
      print_node(fd, car(l));
    }
    else {
      print_node(fd, name);
      fprintf(fd, " = ");
      print_node(fd, car(l));
    }
    fprintf(fd, "\n");
  }
}

/* returns the list of all the possible values that the variable name
   must assume in set */
static node_ptr extract_values(node_ptr name, BddFsm fsm, bdd_ptr set) 
{
  node_ptr vals = Nil;
  node_ptr range = BddFsm_GetVarRange(fsm, name);

  /* possible optimization checking the intersections of supports. If
     no intersection, then return the range as it is: all values are
     possible. */
  for ( ; range != Nil; range = cdr(range)) {
    bdd_ptr  tst;
    node_ptr sv = car(range);
    bdd_ptr  bv = BddFsm_EvalCond(fsm, mk_equal(name, sv));

    tst = bdd_and(dd_manager, set, bv);
    if (bdd_isnot_zero(dd_manager, tst) == 1) {
      vals = cons(sv, vals);
    }
    bdd_free(dd_manager, tst);
    bdd_free(dd_manager, bv);
  }
  /* we reverse the list to better print results. */
  vals = reverse(vals);
  return(vals);
}

/* extracts k element from list l */
static node_ptr get_element(node_ptr l, int k) 
{
  int j;
  node_ptr result;

  assert(l != Nil);
  for(j = 0 ; j < k; j++, l = cdr(l));
  assert(l != Nil);
  result = car(l);
  return(result);
}

static void free_simulation_path(node_ptr path)
{
  node_ptr l;
  for(l = path; l != Nil; l = cdr(l)) {
    /* ((Sbdd . (Ol . Bbdd)) . Abdd) */
    node_ptr s = car(l);
#if 0
    /* car(car(s)) = Sbdd */
    bdd_free(dd_manager, (bdd_ptr)car(car(s)));
    /* cdr(cdr(car(s)) = Bbdd */
    bdd_free(dd_manager, (bdd_ptr)cdr(cdr(car(s))));
#endif
    /* cdr(s) = Abdd */
    bdd_free(dd_manager, (bdd_ptr)cdr(s));
  }
}

static node_ptr mk_equal(node_ptr l, node_ptr r)
{
  return(find_node(EQUAL, l, r));
}

static node_ptr mk_true ARGS((void))
{
  return(find_node(TRUEEXP, Nil, Nil));
}

static node_ptr mk_false ARGS((void))
{
  return(find_node(FALSEEXP, Nil, Nil));
}

static void simulate_print_path_generic(FILE *fd, BddFsm fsm, node_ptr path) 
{
  int step = 0;
  node_ptr l;
  
  if (simulate_int_print_trace()) {
    for(l = path; l != Nil; l = cdr(l), step++) {
      /* ((Sbdd . (Ol . Bbdd)) . Abdd) */
      node_ptr s = car(l);
      /* car(car(s)) = Sbdd */
      bdd_ptr State = (bdd_ptr)car(car(s));
      bdd_ptr input = (bdd_ptr)cdr(s);

      fprintf(fd, "\n==== Begin Step [%6d] =========\n", step);
      fprintf(fd, "==== State: ");
      simulate_print_state(fd, BddFsm_WffCond(fsm, State));
      fprintf(fd, "\n\n");
      /* input is printed out only if different from 0/1 */
      if (bdd_isnot_zero(dd_manager, input) &&
           bdd_isnot_one(dd_manager, input)) {
        fprintf(fd, "==== Inputs: ");
        simulate_print_input(fd, BddFsm_WffCond(fsm, input));
        fprintf(fd, "\n\n");
      }
      fprintf(fd,   "==== End Step   [%6d] ===========\n\n", step);
    }
  }
}

static void simulate_print_path_god(FILE * fd, BddFsm fsm, node_ptr path) {
  fprintf(fd, "=========== Begin Simulation Path Printing ===========\n");
  simulate_print_path_generic(fd, fsm, path);
  fprintf(fd, "=========== End Simulation Path Printing ===========\n");
}  



static void simulate_save_path_generic(FILE *fd, BddFsm fsm, node_ptr path) 
{
  if (simulate_int_save_trace()) {
    node_ptr l;
    int state_no = 1;
    int trace_no = get_trace_number();

    /* we increment trace number for further use */
    inc_trace_number();

    for(l = path; l != Nil; l = cdr(l), state_no++) {
      /* ((Sbdd . (Ol . Bbdd)) . Abdd) */
      node_ptr s = car(l);
      /* car(car(s)) = Sbdd */
      bdd_ptr state = (bdd_ptr)car(car(s));
      bdd_ptr pinput = (bdd_ptr)cdr(s);
      node_ptr label = trace_mk_statelabel(trace_no, state_no);

      /* input is conjoined only if different from 0/1 */
      if (bdd_isnot_zero(dd_manager, pinput) &&
           bdd_isnot_one(dd_manager, pinput)) {
        bdd_and_accumulate(dd_manager, &state, pinput);
      }

      /* we save the bdd since we insert it in the trace hash */
      bdd_ref(state);
      insert_traces_hash(label, (node_ptr)state);

      /* we free partial computations */
      bdd_free(dd_manager, state);
      bdd_free(dd_manager, pinput);
    }
  }
}

static void simulate_save_path_god(FILE * fd, BddFsm fsm, node_ptr path) {
  simulate_save_path_generic(fd, fsm, path);
}  

static void simulatePrintState(FILE * fd, BddFsm fsm, bdd_ptr s)
{
  if (bdd_isnot_zero(dd_manager, s)) {
    node_ptr st = BddFsm_WffCond(fsm, s);
  
    simulate_print_state(fd, st);
  }
}  

static void simulatePrintInput(FILE * fd, BddFsm fsm, bdd_ptr a)
{
  if (bdd_isnot_zero(dd_manager, a)) {
    node_ptr at = BddFsm_WffCond(fsm, a);

    fprintf(fd, "\t");
    print_node(fd, at);
    fprintf(fd, "\n");
  }
}  


static int ask_user_yes_no(const char * msg) 
{
  int c;
  char p = '\0';

  do {
    fprintf(stdout, "\n%s: [y/n] ", msg);
    p = getchar();
    /* to read carriage return */
    if (p != '\n') getchar();
    fflush(stdin);
    c = ((p == 'y') || (p == 'n'));
  } while (((p != EOF) && (p != 1)) && (c != 1));
  return(p == 'y');
}

static int simulate_int_print_trace() 
{
  return(ask_user_yes_no("Would you print the simulated trace?"));
}

static int simulate_int_save_trace() 
{
  return(ask_user_yes_no("Would you save the simulated trace?"));
}

static boolean simulate_continue_simulation() {
  if (ask_user_yes_no("Would you continue simulation?")) {
    return(false);
  }
  return(true);
}

static int is_to_avoid_name(node_ptr name) {
  assert(name != Nil);
  switch(node_get_type(name)) {
  case DOT:
    return is_to_avoid_name(cdr(name));
    break;
  case ATOM:
    if ((strcmp(get_text((string_ptr)car(name)), "_constants")) == 0) {
      return(1);
    }
    else {
      int p, q;
      if (sscanf(get_text((string_ptr)car(name)),
                 "LTL_%d_SPECF_%d", &p, &q) == 2) {
        return(1);
      }
    }
    break;
  default:
    return(0);
    break;
  }
  return(0);
}

static int check_interactive_read(int r, int k, int n) {
  return((((r != EOF) && ((k >= n) || (k < 0)))));
}

static void simulate_print_state(FILE *fd, node_ptr state) {
  if (state == Nil) return;

  switch(node_get_type(state)) {
  case AND:
    simulate_print_state(fd, car(state));
    simulate_print_state(fd, cdr(state));
    break;
  case EQUAL:
    if (is_to_avoid_name(car(state)) == 0) {
      fprintf(fd, "\n\t");
      print_node(fd, state);
    }
    break;
  case ATOM:
  case DOT:
    if (is_to_avoid_name(state) == 0) {
      fprintf(fd, "\n\t");
      print_node(fd, state);
    }
    break;
  default:
    fprintf(fd, "\n\t");
    print_node(fd, state);
    break;
  }
  return;
}

static void simulate_print_input(FILE *fd, node_ptr input) {
  if (input == Nil) return;

  switch(node_get_type(input)) {
  case AND:
    simulate_print_input(fd, car(input));
    simulate_print_input(fd, cdr(input));
    break;
  case EQUAL:
    if (is_to_avoid_name(car(input)) == 0) {
      fprintf(fd, "\n\t");
      print_node(fd, input);
    }
    break;
  case ATOM:
  case DOT:
    if (is_to_avoid_name(input) == 0) {
      fprintf(fd, "\n\t");
      print_node(fd, input);
    }
    break;
  default:
    fprintf(fd, "\n\t");
    print_node(fd, input);
    break;
  }
  return;
}



static void print_trace_state(bdd_ptr s, node_ptr symbols, int changes_only)
{
  node_ptr proc_name;
  node_ptr process_selector_data;
  add_ptr add_state, tmp;
  add_ptr process_selector_enc;
  node_ptr los = symbols;

  if (bdd_isnot_zero(dd_manager, s)) {
    add_state = bdd_to_add(dd_manager, s);
    process_selector_data = lookup_symbol_hash(proc_selector_internal_vname);
    if (process_selector_data != NULL) {
      process_selector_enc = (add_ptr)car(process_selector_data);
      tmp = add_if_then(dd_manager, add_state, process_selector_enc);
      proc_name = add_value(dd_manager, tmp);
      add_free(dd_manager, tmp);

      if ((proc_name != Nil) && (proc_name != sym_intern("main"))) {
        indent_node(nusmv_stdout, "", proc_selector_internal_vname, " = ");
        print_node(nusmv_stdout, proc_name);
      }
    }
    while (los) {
      node_ptr cur_sym_value;
      node_ptr cur_sym = car(los);
      add_ptr cur_sym_vals;

      los = cdr(los);
      if (is_to_avoid_name(cur_sym) == 0) {
        cur_sym_vals = eval(cur_sym, Nil);
        tmp = add_if_then(dd_manager, add_state, cur_sym_vals);
        cur_sym_value = add_value(dd_manager, tmp);
        add_free(dd_manager, tmp);
        add_free(dd_manager, cur_sym_vals);
        if (changes_only) {
          if (cur_sym_value == lookup_print_hash(cur_sym)) continue;
          insert_print_hash(cur_sym, cur_sym_value);
        }
        indent_node(nusmv_stdout, "", cur_sym, " = ");
        print_node(nusmv_stdout, cur_sym_value);
        fprintf(nusmv_stdout, "\n");
      }
    }
    add_free(dd_manager, add_state);
  }
  else {
    indent_node(nusmv_stdout, "", mk_false(), "\n");
  }
}

