/**CFile***********************************************************************

  FileName    [simulatorCmd.c]

  PackageName [simulator]

  Synopsis    [The new simulation engine command interface]

  Description [This file provides the first version of the new
  interactive symulator.]

  SeeAlso     [optional]

  Author      [Marco Roveri]

  Copyright   [This file is part of the ``simulator'' package of NuSMV
  version 2.  Copyright (C) 2003 ITC-irst.

  NuSMV version 2 is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2 of
  the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it> or
  email to <nusmv-users@irst.itc.it>.  Please report bugs to
  <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to
  <nusmv@irst.itc.it>.]

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>

#include "util.h"
#include "utils.h"
#include "node.h"
#include "grammar.h"
#include "dd.h"
#include "bddfsm.h"
#include "compile.h"
#include "dsimulate.h"
#include "opts.h"

static char rcsid[] UTIL_UNUSED = "$Id: simulatorCmd.c,v 1.3 2003/10/09 10:07:54 nusmv Exp $";

int CommandDSimulate ARGS((int argc, char **argv));
int CommandSetDsimOptions ARGS((int argc, char **argv));
int CommandDsimShowTrace  ARGS((int argc, char **argv));

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
Opts_HandlerPtr simopthr = (Opts_HandlerPtr)NULL;
extern cmp_struct_ptr cmps;

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static int UsageDSimulate ARGS((void));
static int UsageSetDsimOptions ARGS((void));
static int UsageDsimShowTrace ARGS((void));

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/
void Simulator_Init(void){
  simopthr = Opts_HandlerAlloc();
  if (simopthr == OPTS_ERROR) {
    fprintf(nusmv_stderr, "Unable to allocate Simulation option handler.\n");
    exit(1);
  }
  DSimulate_RegisterOptions();
  Cmd_CommandAdd("dsimulate", CommandDSimulate, 0);
  Cmd_CommandAdd("dsimulate_set_options", CommandSetDsimOptions);
  Cmd_CommandAdd("dsimulate_show_trace", CommandDsimShowTrace);
}

void Simulator_End(void){
  DSimulate_UnRegisterOptions();
  Opts_HandlerFree(&simopthr);
}


int CommandDSimulate(int argc, char **argv)
{
  /* skipping command name */
  argc--;  argv++;

  /* command line option parsing */
  while(argc > 0) {
    if (strcmp(*argv,"-h") == 0) {
      argv++; argc--;
      return(UsageDSimulate());
    }
    else if (DSimulate_ParseCLOptions(&argv, &argc) == 1) {
      continue;
    }
    else {
      return(UsageDSimulate());
    }   
  }

  if (cmp_struct_get_read_model(cmps) == 0) {
    fprintf(nusmv_stderr,
            "A model must be read before. Use the \"read_model\" command.\n");
    return(1);
  }
  if (cmp_struct_get_flatten_hrc(cmps) == 0) {
    fprintf(nusmv_stderr,
            "The hierarchy must be flattened before. Use the \"flatten_hierarchy\" command.\n");
    return(1);
  }
  if (cmp_struct_get_encode_variables(cmps) == 0) {
    fprintf(nusmv_stderr, "The variables must be built before. Use the \"encode_variables\" command.\n");
    return(1);
  }
  if (cmp_struct_get_build_model(cmps) == 0) {
    fprintf(nusmv_stderr, "A model must be built before. Use the \"build_model\" command.\n");
    return(1);
  }

  DSimulate_DomainSimulate(Prop_MasterGetBddFsm());
  return(0);
}

static int UsageDSimulate(void)
{
  fprintf(nusmv_stderr, "usage: dsimulate [-h] [-simulate_sel_mode <mode>]\n\t\t [-simulate_maxsteps #int]\n\t\t [-simulate_tab_vars_mode <mode>] \n");
  fprintf(nusmv_stderr, "\t -h \t\t\t\tPrints the command usage.\n");
  fprintf(nusmv_stderr, "\t -simulate_sel_mode <mode>\tSpecifies simulation state selection\n\t\t\t\t\t mode: <mode>={rnd|fix|int}.\n");
  fprintf(nusmv_stderr, "\t -simulate_tab_vars_mode <mode>\tSpecifies the selection mode for\n\t\t\t\t\t tableau variables: <mode>={fix|nofix}.\n");
  fprintf(nusmv_stderr, "\t -simulate_maxsteps #int\tSpecifies maximum number of simulation\n\t\t\t\t\t steps.\n");
  return(1);
}



int CommandSetDsimOptions(int argc, char **argv)
{
  int v = 0;
  /* skipping command name */
  argc--;  argv++;

  /* command line option parsing */
  while(argc > 0) {
    if (strcmp(*argv,"-h") == 0) {
      argv++; argc--;
      return(UsageSetDsimOptions());
    }
    if (strcmp(*argv,"-v") == 0) {
      argv++; argc--;
      v = 1;
      continue;
    }
    else if (DSimulate_ParseCLOptions(&argv, &argc) == 1) {
      continue;
    }
    else {
      return(UsageSetDsimOptions());
    }   
  }
  if (v == 1) {
    Opts_PrintAllOptions(simopthr, nusmv_stdout);
  }
  return(0);
}

static int UsageSetDsimOptions(void)
{
  fprintf(nusmv_stderr, "usage: dsimulate_set_options [-h] [-v]\n\t\t [-simulate_sel_mode <mode>]\n\t\t [-simulate_maxsteps #int]\n\t\t [-simulate_tab_vars_mode <mode>]\n");
  fprintf(nusmv_stderr, "\t -h \t\t\t\tPrints the command usage.\n");
  fprintf(nusmv_stderr, "\t -v \t\t\t\tPrints the simualtion options.\n");
  fprintf(nusmv_stderr, "\t -simulate_sel_mode <mode>\tSpecifies simulation state selection\n\t\t\t\t\t mode: <mode>={rnd|fix|int}.\n");
  fprintf(nusmv_stderr, "\t -simulate_maxsteps #int\tSpecifies maximum number of simulation\n\t\t\t\t\t steps.\n");
  fprintf(nusmv_stderr, "\t -simulate_tab_vars_mode <mode>\tSpecifies the selection mode for\n\t\t\t\t\t tableau variables: <mode>={fix|nofix}.\n");
  return(1);
}

int CommandDsimShowTrace(int argc, char **argv) 
{
  char c;
  char *num = NIL(char);
  int verbose = 1;
  int trace = -1;
  int maxt = get_trace_number() -1;

  util_getopt_reset();
  while((c = util_getopt(argc,argv,"hvn:")) != EOF){
    switch(c){
    case 'h': return(UsageDsimShowTrace());
    case 'v':
      verbose = 0;
      break;
    case 'n':
      if (num != NIL(char)) return(UsageDsimShowTrace());
      num = util_strsav(util_optarg);
      break;
    default:
      return(UsageDsimShowTrace());
      break;
    }
  }
  if (num != NIL(char)) {
    char *err_occ[2];

    trace = strtol(num, err_occ, 10);
    if ((strcmp(err_occ[0], "") != 0) ||
        (trace < 1)) {
      fprintf(nusmv_stderr,
              "Error: \"%s\" is not a valid value (must be a positive integer).\n", err_occ[0]);
      return(1);
    }
  }

  if (trace > maxt) {
      fprintf(nusmv_stderr,
              "Error: \"%d\" is not a valid trace number (must be a positive integer < %d ).\n", trace, maxt);
      return(1);
  }

  if (trace != -1) {
    print_tropos_trace(trace, verbose);
  }
  else {
    switch(maxt) {
    case 0:
      fprintf(nusmv_stdout, "There are no traces currently stored.\n");
      break;
    case 1:
      fprintf(nusmv_stdout, "There is 1 trace currently stored.\n");
      break;
    default:
      fprintf(nusmv_stdout, "There are %d traces currently stored.\n", maxt);
      break;
    }
  }
  return(0);
}

static int UsageDsimShowTrace(void)
{
  fprintf(nusmv_stderr, "usage: dsimulate_show_trace [-h] [-v] [-n #nat]\n");
  fprintf(nusmv_stderr, "\t -h \t\tPrints the command usage.\n");
  fprintf(nusmv_stderr, "\t -v \t\tPrints verbosely traces.\n");
  fprintf(nusmv_stderr, "\t -n #nat\tSpecifies the trace to show.\n");
  return(1);
}


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/



